import json
import logging
import os
import re
import traceback
from time import sleep

import hypothesis.strategies as hst
import numpy as np
import pytest
from hypothesis import HealthCheck, given, settings
from numpy.testing import assert_allclose, assert_array_equal
from unittest.mock import patch

import qcodes as qc
from qcodes.dataset.data_set import load_by_id
from qcodes.dataset.descriptions.param_spec import ParamSpecBase
from qcodes.dataset.experiment_container import new_experiment
from qcodes.dataset.legacy_import import import_dat_file
from qcodes.dataset.measurements import Measurement
from qcodes.dataset.sqlite.connection import atomic_transaction
from qcodes.instrument.parameter import (ArrayParameter, Parameter,
                                         expand_setpoints_helper)
from qcodes.tests.common import retry_until_does_not_throw, reset_config_on_exit
# pylint: disable=unused-import
from qcodes.tests.test_station import set_default_station_to_none
from qcodes.utils.validators import Arrays
from qcodes.dataset.export_config import DataExportType


def test_log_messages(caplog, meas_with_registered_param):
    caplog.set_level(logging.INFO)

    with meas_with_registered_param.run():
        pass

    assert "Set the run_timestamp of run_id" in caplog.text
    assert "Starting measurement with guid" in caplog.text
    assert "Finished measurement with guid" in caplog.text


def test_log_includes_extra_info(caplog, meas_with_registered_param):
    caplog.set_level(logging.INFO)
    meas_with_registered_param._extra_log_info = "some extra info"
    with meas_with_registered_param.run():
        pass

    assert "some extra info" in caplog.text


def test_register_parameter_numbers(DAC, DMM):
    """
    Test the registration of scalar QCoDeS parameters
    """

    parameters = [DAC.ch1, DAC.ch2, DMM.v1, DMM.v2]
    not_parameters = ['', 'Parameter', 0, 1.1, Measurement]

    meas = Measurement()

    for not_a_parameter in not_parameters:
        with pytest.raises(ValueError):
            meas.register_parameter(not_a_parameter)

    my_param = DAC.ch1
    meas.register_parameter(my_param)

    assert len(meas.parameters) == 1
    paramspec = meas.parameters[str(my_param)]
    assert paramspec.name == str(my_param)
    assert paramspec.label == my_param.label
    assert paramspec.unit == my_param.unit
    assert paramspec.type == 'numeric'

    # we allow the registration of the EXACT same parameter twice...
    meas.register_parameter(my_param)
    # ... but not a different parameter with a new name
    attrs = ['label', 'unit']
    vals = ['new label', 'new unit']
    for attr, val in zip(attrs, vals):
        old_val = getattr(my_param, attr)
        setattr(my_param, attr, val)
        match = re.escape("Parameter already registered in this Measurement.")
        with pytest.raises(ValueError, match=match):
            meas.register_parameter(my_param)
        setattr(my_param, attr, old_val)

    assert len(meas.parameters) == 1
    paramspec = meas.parameters[str(my_param)]
    assert paramspec.name == str(my_param)
    assert paramspec.label == my_param.label
    assert paramspec.unit == my_param.unit
    assert paramspec.type == 'numeric'

    for parameter in parameters:
        with pytest.raises(ValueError):
            meas.register_parameter(my_param, setpoints=(parameter,))
        with pytest.raises(ValueError):
            meas.register_parameter(my_param, basis=(parameter,))

    meas.register_parameter(DAC.ch2)
    meas.register_parameter(DMM.v1)
    meas.register_parameter(DMM.v2)
    meas.unregister_parameter(my_param)
    meas.register_parameter(my_param, basis=(DAC.ch2,),
                            setpoints=(DMM.v1, DMM.v2))

    assert set(meas.parameters.keys()) == {str(DAC.ch2),
                                               str(DMM.v1), str(DMM.v2),
                                               str(my_param)}
    paramspec = meas.parameters[str(my_param)]
    assert paramspec.name == str(my_param)

    meas = Measurement()

    meas.register_parameter(DAC.ch1)
    meas.register_parameter(DAC.ch2, setpoints=(DAC.ch1,))
    with pytest.raises(ValueError):
        meas.register_parameter(DMM.v1, setpoints=(DAC.ch2,))


def test_register_custom_parameter(DAC):
    """
    Test the registration of custom parameters
    """
    meas = Measurement()

    name = 'V_modified'
    unit = 'V^2'
    label = 'square of the voltage'

    meas.register_custom_parameter(name, label, unit)

    assert len(meas.parameters) == 1
    assert isinstance(meas.parameters[name], ParamSpecBase)
    assert meas.parameters[name].unit == unit
    assert meas.parameters[name].label == label
    assert meas.parameters[name].type == 'numeric'

    newunit = 'V^3'
    newlabel = 'cube of the voltage'

    meas.unregister_parameter(name)
    meas.register_custom_parameter(name, newlabel, newunit)

    assert len(meas.parameters) == 1
    assert isinstance(meas.parameters[name], ParamSpecBase)
    assert meas.parameters[name].unit == newunit
    assert meas.parameters[name].label == newlabel

    with pytest.raises(ValueError):
        meas.register_custom_parameter(name, label, unit,
                                       setpoints=(DAC.ch1,))
    with pytest.raises(ValueError):
        meas.register_custom_parameter(name, label, unit,
                                       basis=(DAC.ch2,))

    meas.register_parameter(DAC.ch1)
    meas.register_parameter(DAC.ch2)
    meas.register_custom_parameter('strange_dac')

    meas.unregister_parameter(name)
    meas.register_custom_parameter(name, label, unit,
                                   setpoints=(DAC.ch1, str(DAC.ch2)),
                                   basis=('strange_dac',))

    assert len(meas.parameters) == 4
    parspec = meas.parameters[name]

    with pytest.raises(ValueError):
        meas.register_custom_parameter('double dependence',
                                       'label', 'unit', setpoints=(name,))


def test_unregister_parameter(DAC, DMM):
    """
    Test the unregistering of parameters.
    """

    DAC.add_parameter('impedance',
                      get_cmd=lambda: 5)

    meas = Measurement()

    meas.register_parameter(DAC.ch2)
    meas.register_parameter(DMM.v1)
    meas.register_parameter(DMM.v2)
    meas.register_parameter(DAC.ch1, basis=(DMM.v1, DMM.v2),
                            setpoints=(DAC.ch2,))

    with pytest.raises(ValueError):
        meas.unregister_parameter(DAC.ch2)
    with pytest.raises(ValueError):
        meas.unregister_parameter(str(DAC.ch2))
    with pytest.raises(ValueError):
        meas.unregister_parameter(DMM.v1)
    with pytest.raises(ValueError):
        meas.unregister_parameter(DMM.v2)

    meas.unregister_parameter(DAC.ch1)
    assert set(meas.parameters.keys()) == {str(DAC.ch2), str(DMM.v1),
                                               str(DMM.v2)}

    meas.unregister_parameter(DAC.ch2)
    assert set(meas.parameters.keys()) == {str(DMM.v1), str(DMM.v2)}

    not_parameters = [DAC, DMM, 0.0, 1]
    for notparam in not_parameters:
        with pytest.raises(ValueError):
            meas.unregister_parameter(notparam)

    # unregistering something not registered should silently "succeed"
    meas.unregister_parameter('totes_not_registered')
    meas.unregister_parameter(DAC.ch2)
    meas.unregister_parameter(DAC.ch2)


@pytest.mark.usefixtures("experiment")
@pytest.mark.parametrize("bg_writing", [True, False])
def test_mixing_array_and_numeric(DAC, bg_writing):
    """
    Test that mixing array and numeric types is okay
    """
    meas = Measurement()
    meas.register_parameter(DAC.ch1, paramtype='numeric')
    meas.register_parameter(DAC.ch2, paramtype='array')

    with meas.run(write_in_background=bg_writing) as datasaver:
        datasaver.add_result((DAC.ch1, np.array([DAC.ch1(), DAC.ch1()])),
                             (DAC.ch2, np.array([DAC.ch2(), DAC.ch1()])))


def test_measurement_name_default(experiment, DAC, DMM):
    fmt = experiment.format_string
    exp_id = experiment.exp_id

    default_name = 'results'

    meas = Measurement()
    assert meas.name == ''

    meas.register_parameter(DAC.ch1)
    meas.register_parameter(DMM.v1, setpoints=[DAC.ch1])

    with meas.run() as datasaver:
        run_id = datasaver.run_id
        expected_name = fmt.format(default_name, exp_id, run_id)
        assert datasaver.dataset.table_name == expected_name
        assert datasaver.dataset.name == default_name


def test_measurement_name_changed_via_attribute(experiment, DAC, DMM):
    fmt = experiment.format_string
    exp_id = experiment.exp_id

    name = 'yolo'

    meas = Measurement()
    meas.name = name

    meas.register_parameter(DAC.ch1)
    meas.register_parameter(DMM.v1, setpoints=[DAC.ch1])

    with meas.run() as datasaver:
        run_id = datasaver.run_id
        expected_name = fmt.format('results', exp_id, run_id)
        assert datasaver.dataset.table_name == expected_name
        assert datasaver.dataset.name == name


def test_measurement_name_set_as_argument(experiment, DAC, DMM):
    fmt = experiment.format_string
    exp_id = experiment.exp_id

    name = 'yolo'

    meas = Measurement(name=name, exp=experiment)

    meas.register_parameter(DAC.ch1)
    meas.register_parameter(DMM.v1, setpoints=[DAC.ch1])

    with meas.run() as datasaver:
        run_id = datasaver.run_id
        expected_name = fmt.format('results', exp_id, run_id)
        assert datasaver.dataset.table_name == expected_name
        assert datasaver.dataset.name == name


@settings(deadline=None)
@given(wp=hst.one_of(hst.integers(), hst.floats(allow_nan=False),
                     hst.text()))
@pytest.mark.usefixtures("empty_temp_db")
def test_setting_write_period(wp):
    new_experiment('firstexp', sample_name='no sample')
    meas = Measurement()
    meas.register_custom_parameter(name='dummy')

    if isinstance(wp, str):
        with pytest.raises(ValueError):
            meas.write_period = wp
    elif wp < 1e-3:
        with pytest.raises(ValueError):
            meas.write_period = wp
    else:
        meas.write_period = wp
        assert meas._write_period == float(wp)

        with meas.run() as datasaver:
            assert datasaver.write_period == float(wp)


@settings(deadline=None)
@given(wp=hst.one_of(hst.integers(), hst.floats(allow_nan=False),
                     hst.text()))
@pytest.mark.usefixtures("experiment")
def test_setting_write_period_from_config(wp):
    with reset_config_on_exit():
        qc.config.dataset.write_period = wp

        if isinstance(wp, str):
            with pytest.raises(ValueError):
                Measurement()
        elif wp < 1e-3:
            with pytest.raises(ValueError):
                Measurement()
        else:
            meas = Measurement()
            assert meas.write_period == float(wp)
            meas.register_custom_parameter(name='dummy')
            with meas.run() as datasaver:
                assert datasaver.write_period == float(wp)


@pytest.mark.parametrize("write_in_background", [True, False])
@pytest.mark.usefixtures("experiment")
def test_setting_write_in_background_from_config(write_in_background):
    with reset_config_on_exit():
        qc.config.dataset.write_in_background = write_in_background

        meas = Measurement()
        meas.register_custom_parameter(name='dummy')
        with meas.run() as datasaver:
            assert datasaver.dataset._writer_status.write_in_background is write_in_background


@pytest.mark.usefixtures("experiment")
def test_method_chaining(DAC):
    meas = (
        Measurement()
            .register_parameter(DAC.ch1)
            .register_custom_parameter(name='freqax',
                                       label='Frequency axis',
                                       unit='Hz')
            .add_before_run((lambda: None), ())
            .add_after_run((lambda: None), ())
            .add_subscriber((lambda values, idx, state: None), state=[])
    )


@pytest.mark.usefixtures("experiment")
@settings(deadline=None, suppress_health_check=(HealthCheck.function_scoped_fixture,))
@given(words=hst.lists(elements=hst.text(), min_size=4, max_size=10))
def test_enter_and_exit_actions(DAC, words):
    # we use a list to check that the functions executed
    # in the correct order

    def action(lst, word):
        lst.append(word)

    meas = Measurement()
    meas.register_parameter(DAC.ch1)

    testlist = []

    splitpoint = round(len(words) / 2)
    for n in range(splitpoint):
        meas.add_before_run(action, (testlist, words[n]))
    for m in range(splitpoint, len(words)):
        meas.add_after_run(action, (testlist, words[m]))

    assert len(meas.enteractions) == splitpoint
    assert len(meas.exitactions) == len(words) - splitpoint

    with meas.run() as _:
        assert testlist == words[:splitpoint]

    assert testlist == words

    meas = Measurement()

    with pytest.raises(ValueError):
        meas.add_before_run(action, 'no list!')
    with pytest.raises(ValueError):
        meas.add_after_run(action, testlist)


def test_subscriptions(experiment, DAC, DMM):
    """
    Test that subscribers are called at the moment the data is flushed to
    database

    Note that for the purpose of this test, flush_data_to_database method is
    called explicitly instead of waiting for the data to be flushed
    automatically after the write_period passes after a add_result call.
    """

    def collect_all_results(results, length, state):
        """
        Updates the *state* to contain all the *results* acquired
        during the experiment run
        """
        # Due to the fact that by default subscribers only hold 1 data value
        # in their internal queue, this assignment should work (i.e. not
        # overwrite values in the "state" object) assuming that at the start
        # of the experiment both the dataset and the *state* objects have
        # the same length.
        state[length] = results

    def collect_values_larger_than_7(results, length, state):
        """
        Appends to the *state* only the values from *results*
        that are larger than 7
        """
        for result_tuple in results:
            state += [value for value in result_tuple if value > 7]

    meas = Measurement(exp=experiment)
    meas.register_parameter(DAC.ch1)
    meas.register_parameter(DMM.v1, setpoints=(DAC.ch1,))

    # key is the number of the result tuple,
    # value is the result tuple itself
    all_results_dict = {}
    values_larger_than_7 = []

    meas.add_subscriber(collect_all_results, state=all_results_dict)
    assert len(meas.subscribers) == 1
    meas.add_subscriber(collect_values_larger_than_7,
                        state=values_larger_than_7)
    assert len(meas.subscribers) == 2

    meas.write_period = 0.2

    with meas.run() as datasaver:

        # Assert that the measurement, runner, and datasaver
        # have added subscribers to the dataset
        assert len(datasaver._dataset.subscribers) == 2

        assert all_results_dict == {}
        assert values_larger_than_7 == []

        dac_vals_and_dmm_vals = list(zip(range(5), range(3, 8)))
        values_larger_than_7__expected = []

        for num in range(5):
            (dac_val, dmm_val) = dac_vals_and_dmm_vals[num]
            values_larger_than_7__expected += \
                [val for val in (dac_val, dmm_val) if val > 7]

            datasaver.add_result((DAC.ch1, dac_val), (DMM.v1, dmm_val))

            # Ensure that data is flushed to the database despite the write
            # period, so that the database triggers are executed, which in turn
            # add data to the queues within the subscribers
            datasaver.flush_data_to_database()

            # In order to make this test deterministic, we need to ensure that
            # just enough time has passed between the moment the data is
            # flushed to database and the "state" object (that is passed to
            # subscriber constructor) has been updated by the corresponding
            # subscriber's callback function. At the moment, there is no robust
            # way to ensure this. The reason is that the subscribers have
            # internal queue which is populated via a trigger call from the SQL
            # database, hence from this "main" thread it is difficult to say
            # whether the queue is empty because the subscriber callbacks have
            # already been executed or because the triggers of the SQL database
            # has not been executed yet.
            #
            # In order to overcome this problem, a special decorator is used to
            # wrap the assertions. This is going to ensure that some time is
            # given to the Subscriber threads to finish exhausting the queue.
            @retry_until_does_not_throw(
                exception_class_to_expect=AssertionError, delay=0.5, tries=20)
            def assert_states_updated_from_callbacks():
                assert values_larger_than_7 == values_larger_than_7__expected
                assert list(all_results_dict.keys()) == \
                       [result_index for result_index in range(1, num + 1 + 1)]

            assert_states_updated_from_callbacks()

    # Ensure that after exiting the "run()" context,
    # all subscribers get unsubscribed from the dataset
    assert len(datasaver._dataset.subscribers) == 0

    # Ensure that the triggers for each subscriber
    # have been removed from the database
    get_triggers_sql = "SELECT * FROM sqlite_master WHERE TYPE = 'trigger';"
    triggers = atomic_transaction(
        datasaver._dataset.conn, get_triggers_sql).fetchall()
    assert len(triggers) == 0


def test_subscribers_called_at_exiting_context_if_queue_is_not_empty(experiment,
                                                                     DAC):
    """
    Upon quitting the "run()" context, verify that in case the queue is
    not empty, the subscriber's callback is still called on that data.
    This situation is created by setting the minimum length of the queue
    to a number that is larger than the number of value written to the dataset.
    """

    def collect_x_vals(results, length, state):
        """
        Collects first elements of results tuples in *state*
        """
        index_of_x = 0
        state += [res[index_of_x] for res in results]

    meas = Measurement(exp=experiment)
    meas.register_parameter(DAC.ch1)

    collected_x_vals = []

    meas.add_subscriber(collect_x_vals, state=collected_x_vals)

    given_x_vals = [0, 1, 2, 3]

    with meas.run() as datasaver:
        # Set the minimum queue size of the subscriber to more that
        # the total number of values being added to the dataset;
        # this way the subscriber callback is not called before
        # we exit the "run()" context.
        subscriber = list(datasaver.dataset.subscribers.values())[0]
        subscriber.min_queue_length = int(len(given_x_vals) + 1)

        for x in given_x_vals:
            datasaver.add_result((DAC.ch1, x))
            # Verify that the subscriber callback is not called yet
            assert collected_x_vals == []

    # Verify that the subscriber callback is finally called
    assert collected_x_vals == given_x_vals


@pytest.mark.serial
@pytest.mark.flaky(reruns=5)
@settings(deadline=None, max_examples=25,
          suppress_health_check=(HealthCheck.function_scoped_fixture,))
@given(N=hst.integers(min_value=2000, max_value=3000))
def test_subscribers_called_for_all_data_points(experiment, DAC, DMM, N):
    def sub_get_x_vals(results, length, state):
        """
        A list of all x values
        """
        state += [res[0] for res in results]

    def sub_get_y_vals(results, length, state):
        """
        A list of all y values
        """
        state += [res[1] for res in results]

    meas = Measurement(exp=experiment)
    meas.register_parameter(DAC.ch1)
    meas.register_parameter(DMM.v1, setpoints=(DAC.ch1,))
    xvals = []
    yvals = []

    meas.add_subscriber(sub_get_x_vals, state=xvals)
    meas.add_subscriber(sub_get_y_vals, state=yvals)

    given_xvals = range(N)
    given_yvals = range(1, N + 1)

    with meas.run() as datasaver:
        for x, y in zip(given_xvals, given_yvals):
            datasaver.add_result((DAC.ch1, x), (DMM.v1, y))

    assert xvals == list(given_xvals)
    assert yvals == list(given_yvals)


# There is no way around it: this test is slow. We test that write_period
# works and hence we must wait for some time to elapse. Sorry.
@settings(max_examples=5, deadline=None,
          suppress_health_check=(HealthCheck.function_scoped_fixture,))
@given(breakpoint=hst.integers(min_value=1, max_value=19),
       write_period=hst.floats(min_value=0.1, max_value=1.5),
       set_values=hst.lists(elements=hst.floats(), min_size=20, max_size=20),
       get_values=hst.lists(elements=hst.floats(), min_size=20, max_size=20))
@pytest.mark.usefixtures('set_default_station_to_none')
def test_datasaver_scalars(experiment, DAC, DMM, set_values, get_values,
                           breakpoint, write_period):
    no_of_runs = len(experiment)

    station = qc.Station(DAC, DMM)

    meas = Measurement(station=station)
    meas.write_period = write_period

    assert meas.write_period == write_period

    meas.register_parameter(DAC.ch1)
    meas.register_parameter(DMM.v1, setpoints=(DAC.ch1,))

    with meas.run() as datasaver:
        for set_v, get_v in zip(set_values[:breakpoint],
                                get_values[:breakpoint]):
            datasaver.add_result((DAC.ch1, set_v), (DMM.v1, get_v))

        assert datasaver._dataset.number_of_results == 0
        sleep(write_period * 1.1)
        datasaver.add_result((DAC.ch1, set_values[breakpoint]),
                             (DMM.v1, get_values[breakpoint]))
        assert datasaver.points_written == breakpoint + 1

    assert datasaver.run_id == no_of_runs + 1

    with meas.run() as datasaver:
        with pytest.raises(ValueError):
            datasaver.add_result((DAC.ch2, 1), (DAC.ch2, 2))
        with pytest.raises(ValueError):
            datasaver.add_result((DMM.v1, 0))

    # More assertions of setpoints, labels and units in the DB!


@pytest.mark.usefixtures('set_default_station_to_none')
def test_datasaver_inst_metadata(experiment, DAC_with_metadata, DMM):
    """
    Check that additional instrument metadata is captured into the dataset snapshot
    """

    station = qc.Station(DAC_with_metadata, DMM)

    meas = Measurement(station=station)
    meas.register_parameter(DAC_with_metadata.ch1)
    meas.register_parameter(DMM.v1, setpoints=(DAC_with_metadata.ch1,))

    with meas.run() as datasaver:
        for set_v in range(10):
            DAC_with_metadata.ch1.set(set_v)
            datasaver.add_result((DAC_with_metadata.ch1, set_v), (DMM.v1, DMM.v1.get()))
    station_snapshot = datasaver.dataset.snapshot['station']
    assert station_snapshot['instruments']['dummy_dac']['metadata'] == {"dac": "metadata"}


def test_exception_happened_during_measurement_is_stored_in_dataset_metadata(
        experiment):
    meas = Measurement()
    meas.register_custom_parameter(name='nodata')

    class SomeMeasurementException(Exception):
        pass

    # `pytest.raises`` is used here instead of custom try-except for convenience
    with pytest.raises(SomeMeasurementException, match='foo') as e:

        with meas.run() as datasaver:
            dataset = datasaver.dataset

            raise SomeMeasurementException('foo')

    metadata = dataset.metadata
    assert "measurement_exception" in metadata

    expected_exception_string = "".join(
        traceback.format_exception(e.type, e.value, e.tb))
    exception_string = metadata["measurement_exception"]
    assert exception_string == expected_exception_string


@pytest.mark.parametrize("bg_writing", [True, False])
@settings(max_examples=10, deadline=None)
@given(N=hst.integers(min_value=2, max_value=500))
@pytest.mark.usefixtures("empty_temp_db")
def test_datasaver_arrays_lists_tuples(bg_writing, N):
    new_experiment('firstexp', sample_name='no sample')

    meas = Measurement()

    meas.register_custom_parameter(name='freqax',
                                   label='Frequency axis',
                                   unit='Hz')
    meas.register_custom_parameter(name='signal',
                                   label='qubit signal',
                                   unit='Majorana number',
                                   setpoints=('freqax',))

    with meas.run(write_in_background=bg_writing) as datasaver:
        freqax = np.linspace(1e6, 2e6, N)
        signal = np.random.randn(N)

        datasaver.add_result(('freqax', freqax), ('signal', signal))

    assert datasaver.points_written == N
    assert not(datasaver.dataset.conn.atomic_in_progress)

    with meas.run(write_in_background=bg_writing) as datasaver:
        freqax = np.linspace(1e6, 2e6, N)
        signal = np.random.randn(N - 1)

        with pytest.raises(ValueError):
            datasaver.add_result(('freqax', freqax), ('signal', signal))

    meas.register_custom_parameter(name='gate_voltage',
                                   label='Gate tuning potential',
                                   unit='V')
    meas.unregister_parameter('signal')
    meas.register_custom_parameter(name='signal',
                                   label='qubit signal',
                                   unit='Majorana flux',
                                   setpoints=('freqax', 'gate_voltage'))

    # save arrays
    with meas.run(write_in_background=bg_writing) as datasaver:
        freqax = np.linspace(1e6, 2e6, N)
        signal = np.random.randn(N)

        datasaver.add_result(('freqax', freqax),
                             ('signal', signal),
                             ('gate_voltage', 0))

    assert datasaver.points_written == N
    assert not(datasaver.dataset.conn.atomic_in_progress)

    # save lists
    with meas.run(write_in_background=bg_writing) as datasaver:
        freqax = list(np.linspace(1e6, 2e6, N))
        signal = list(np.random.randn(N))

        datasaver.add_result(('freqax', freqax),
                             ('signal', signal),
                             ('gate_voltage', 0))

    assert datasaver.points_written == N

    # save tuples
    with meas.run(write_in_background=bg_writing) as datasaver:
        freqax = tuple(np.linspace(1e6, 2e6, N))
        signal = tuple(np.random.randn(N))

        datasaver.add_result(('freqax', freqax),
                             ('signal', signal),
                             ('gate_voltage', 0))

    assert datasaver.points_written == N


@pytest.mark.parametrize("bg_writing", [True, False])
@settings(max_examples=10, deadline=None)
@given(N=hst.integers(min_value=2, max_value=500))
@pytest.mark.usefixtures("empty_temp_db")
def test_datasaver_numeric_and_array_paramtype(bg_writing, N):
    """
    Test saving one parameter with 'numeric' paramtype and one parameter with
    'array' paramtype
    """
    new_experiment('firstexp', sample_name='no sample')

    meas = Measurement()

    meas.register_custom_parameter(name='numeric_1',
                                   label='Magnetic field',
                                   unit='T',
                                   paramtype='numeric')
    meas.register_custom_parameter(name='array_1',
                                   label='Alazar signal',
                                   unit='V',
                                   paramtype='array',
                                   setpoints=('numeric_1',))

    signal = np.random.randn(113)

    with meas.run(bg_writing) as datasaver:
        datasaver.add_result(('numeric_1', 3.75), ('array_1', signal))

    assert datasaver.points_written == 1

    data = datasaver.dataset.get_parameter_data(
        *datasaver.dataset.parameters.split(','))
    assert (data['numeric_1']['numeric_1'] == np.array([3.75])).all()
    assert np.allclose(data['array_1']['array_1'], signal)


@pytest.mark.parametrize("bg_writing", [True, False])
@pytest.mark.usefixtures("empty_temp_db")
def test_datasaver_numeric_after_array_paramtype(bg_writing):
    """
    Test that passing values for 'array' parameter in `add_result` before
    passing values for 'numeric' parameter works.
    """
    new_experiment('firstexp', sample_name='no sample')

    meas = Measurement()

    meas.register_custom_parameter(name='numeric_1',
                                   label='Magnetic field',
                                   unit='T',
                                   paramtype='numeric')
    meas.register_custom_parameter(name='array_1',
                                   label='Alazar signal',
                                   unit='V',
                                   paramtype='array',
                                   setpoints=('numeric_1',))

    signal = np.random.randn(113)

    with meas.run(write_in_background=bg_writing) as datasaver:
        # it is important that first comes the 'array' data and then 'numeric'
        datasaver.add_result(('array_1', signal), ('numeric_1', 3.75))

    assert datasaver.points_written == 1

    data = datasaver.dataset.get_parameter_data(
        *datasaver.dataset.parameters.split(','))
    assert (data['numeric_1']['numeric_1'] == np.array([3.75])).all()
    assert np.allclose(data['array_1']['array_1'], signal)


@pytest.mark.parametrize("bg_writing", [True, False])
@pytest.mark.usefixtures("experiment")
def test_datasaver_foul_input(bg_writing):
    meas = Measurement()

    meas.register_custom_parameter('foul',
                                   label='something unnatural',
                                   unit='Fahrenheit')

    foul_stuff = [qc.Parameter('foul'), {1, 2, 3}]

    with meas.run(bg_writing) as datasaver:
        for ft in foul_stuff:
            with pytest.raises(ValueError):
                datasaver.add_result(('foul', ft))


@settings(max_examples=10, deadline=None)
@given(N=hst.integers(min_value=2, max_value=500))
@pytest.mark.usefixtures("empty_temp_db")
@pytest.mark.parametrize("bg_writing", [True, False])
@pytest.mark.parametrize("storage_type", ['numeric', 'array'])
def test_datasaver_unsized_arrays(N, storage_type, bg_writing):
    new_experiment('firstexp', sample_name='no sample')

    meas = Measurement()

    meas.register_custom_parameter(name='freqax',
                                   label='Frequency axis',
                                   unit='Hz',
                                   paramtype=storage_type)
    meas.register_custom_parameter(name='signal',
                                   label='qubit signal',
                                   unit='Majorana number',
                                   setpoints=('freqax',),
                                   paramtype=storage_type)
    # note that np.array(some_number) is not the same as the number
    # its also not an array with a shape. Check here that we handle it
    # correctly
    with meas.run(write_in_background=bg_writing) as datasaver:
        freqax = np.linspace(1e6, 2e6, N)
        np.random.seed(0)
        signal = np.random.randn(N)
        for i in range(N):
            myfreq = np.array(freqax[i])
            assert myfreq.shape == ()
            mysignal = np.array(signal[i])
            assert mysignal.shape == ()
            datasaver.add_result(('freqax', myfreq), ('signal', mysignal))

    assert datasaver.points_written == N
    loaded_data = datasaver.dataset.get_parameter_data()['signal']

    np.random.seed(0)
    expected_signal = np.random.randn(N)
    expected_freqax = np.linspace(1e6, 2e6, N)

    if storage_type == 'array':
        expected_freqax = expected_freqax.reshape((N, 1))
        expected_signal = expected_signal.reshape((N, 1))

    assert_allclose(loaded_data['freqax'], expected_freqax)
    assert_allclose(loaded_data['signal'], expected_signal)


@settings(max_examples=5, deadline=None,
          suppress_health_check=(HealthCheck.function_scoped_fixture,))
@given(N=hst.integers(min_value=5, max_value=6),
       M=hst.integers(min_value=4, max_value=5),
       seed=hst.integers(min_value=0, max_value=np.iinfo(np.uint32).max))
@pytest.mark.usefixtures("experiment")
@pytest.mark.parametrize("bg_writing", [True, False])
@pytest.mark.parametrize("param_type", ['np_array', 'tuple', 'list'])
@pytest.mark.parametrize("storage_type", ['numeric', 'array'])
def test_datasaver_arrayparams(SpectrumAnalyzer, DAC, N, M,
                               param_type, storage_type,
                               seed, bg_writing):
    """
    test that data is stored correctly for array parameters that
    return numpy arrays, lists and tuples. Stored both as arrays and
    numeric
    """

    if param_type == 'list':
        spectrum = SpectrumAnalyzer.listspectrum
        spectrum_name = 'dummy_SA_listspectrum'
    elif param_type == 'tuple':
        spectrum = SpectrumAnalyzer.tuplespectrum
        spectrum_name = 'dummy_SA_tuplespectrum'
    elif param_type == 'np_array':
        spectrum = SpectrumAnalyzer.spectrum
        spectrum_name = 'dummy_SA_spectrum'
    else:
        raise RuntimeError("Invalid storage_type")

    meas = Measurement()

    meas.register_parameter(spectrum, paramtype=storage_type)

    assert len(meas.parameters) == 2
    setpoint_paramspec = meas.parameters['dummy_SA_Frequency']
    spectrum_paramspec = meas.parameters[str(spectrum)]
    assert setpoint_paramspec in meas._interdeps.dependencies[spectrum_paramspec]
    assert meas.parameters[str(spectrum)].type == storage_type
    assert meas.parameters['dummy_SA_Frequency'].type == storage_type

    # Now for a real measurement

    meas = Measurement()

    meas.register_parameter(DAC.ch1)
    meas.register_parameter(spectrum, setpoints=[DAC.ch1], paramtype=storage_type)

    assert len(meas.parameters) == 3

    spectrum.npts = M

    np.random.seed(seed)
    with meas.run(write_in_background=bg_writing) as datasaver:
        for set_v in np.linspace(0, 0.01, N):
            datasaver.add_result((DAC.ch1, set_v),
                                 (spectrum, spectrum.get()))

    if storage_type == 'numeric':
        assert datasaver.points_written == N * M
    elif storage_type == 'array':
        assert datasaver.points_written == N

    np.random.seed(seed)
    expected_dac_data = np.repeat(np.linspace(0, 0.01, N), M)
    expected_freq_axis = np.tile(spectrum.setpoints[0], N)
    expected_output = np.array([spectrum.get() for _ in range(N)]).reshape(
        N * M)

    if storage_type == 'array':
        expected_dac_data = expected_dac_data.reshape(N, M)
        expected_freq_axis = expected_freq_axis.reshape(N, M)
        expected_output = expected_output.reshape(N, M)

    data = datasaver.dataset.get_parameter_data()[spectrum_name]

    assert_allclose(data['dummy_dac_ch1'], expected_dac_data)
    assert_allclose(data['dummy_SA_Frequency'], expected_freq_axis)
    assert_allclose(data[spectrum_name], expected_output)


@settings(max_examples=5, deadline=None,
          suppress_health_check=(HealthCheck.function_scoped_fixture,))
@given(N=hst.integers(min_value=5, max_value=500))
@pytest.mark.parametrize("bg_writing", [True, False])
@pytest.mark.parametrize("storage_type", ['numeric', 'array'])
@pytest.mark.usefixtures("experiment")
def test_datasaver_array_parameters_channel(channel_array_instrument,
                                            DAC, N, storage_type,
                                            bg_writing):
    meas = Measurement()

    array_param = channel_array_instrument.A.dummy_array_parameter

    meas.register_parameter(array_param, paramtype=storage_type)

    assert len(meas.parameters) == 2
    dependency_name = 'dummy_channel_inst_ChanA_array_setpoint_param_this_setpoint'
    dep_paramspec = meas.parameters[dependency_name]
    array_paramspec = meas.parameters[str(array_param)]
    assert dep_paramspec in meas._interdeps.dependencies[array_paramspec]
    assert meas.parameters[str(array_param)].type == storage_type
    assert meas.parameters[dependency_name].type == storage_type

    # Now for a real measurement

    meas = Measurement()

    meas.register_parameter(DAC.ch1)
    meas.register_parameter(array_param, setpoints=[DAC.ch1], paramtype=storage_type)

    assert len(meas.parameters) == 3

    M = array_param.shape[0]

    with meas.run(write_in_background=bg_writing) as datasaver:
        for set_v in np.linspace(0, 0.01, N):
            datasaver.add_result((DAC.ch1, set_v),
                                 (array_param, array_param.get()))
    if storage_type == 'numeric':
        n_points_written_expected = N * M
    elif storage_type == 'array':
        n_points_written_expected = N

    assert datasaver.points_written == n_points_written_expected

    expected_params = ('dummy_dac_ch1',
                       dependency_name,
                       'dummy_channel_inst_ChanA_dummy_array_parameter')
    ds = load_by_id(datasaver.run_id)
    loaded_data = ds.get_parameter_data()['dummy_channel_inst_ChanA_dummy_array_parameter']
    for param in expected_params:
        if storage_type == 'array':
            expected_shape = (N, M)
        else:
            expected_shape = (N*M, )
        assert loaded_data[param].shape == expected_shape


@settings(max_examples=5, deadline=None,
          suppress_health_check=(HealthCheck.function_scoped_fixture,))
@given(n=hst.integers(min_value=5, max_value=500))
@pytest.mark.parametrize("bg_writing", [True, False])
@pytest.mark.parametrize("storage_type", ['numeric', 'array'])
@pytest.mark.usefixtures("experiment")
def test_datasaver_parameter_with_setpoints(channel_array_instrument,
                                            DAC, n, storage_type,
                                            bg_writing):
    random_seed = 1
    chan = channel_array_instrument.A
    param = chan.dummy_parameter_with_setpoints
    chan.dummy_n_points(n)
    chan.dummy_start(0)
    chan.dummy_stop(100)
    meas = Measurement()
    meas.register_parameter(param, paramtype=storage_type)

    assert len(meas.parameters) == 2
    dependency_name = 'dummy_channel_inst_ChanA_dummy_sp_axis'

    dep_ps = meas.parameters[dependency_name]
    param_ps = meas.parameters[str(param)]

    assert dep_ps in meas._interdeps.dependencies[param_ps]
    assert meas.parameters[str(param)].type == storage_type
    assert meas.parameters[dependency_name].type == storage_type

    # Now for a real measurement
    with meas.run(write_in_background=bg_writing) as datasaver:
        # we seed the random number generator
        # so we can test that we get the expected numbers
        np.random.seed(random_seed)
        datasaver.add_result((param, param.get()))
    if storage_type == 'numeric':
        expected_points_written = n
    elif storage_type == 'array':
        expected_points_written = 1

    assert datasaver.points_written == expected_points_written

    expected_params = (dependency_name,
                       'dummy_channel_inst_ChanA_dummy_parameter_with_setpoints')
    ds = load_by_id(datasaver.run_id)
    loaded_data = ds.get_parameter_data()
    for param in expected_params:
        data = loaded_data['dummy_channel_inst_ChanA_dummy_parameter_with_setpoints'][param]
        if storage_type == 'array':
            assert data.shape == (expected_points_written, n)
        else:
            assert data.shape == (expected_points_written,)
    assert len(loaded_data) == 1

    subdata = loaded_data[
        'dummy_channel_inst_ChanA_dummy_parameter_with_setpoints']

    expected_dep_data = np.linspace(chan.dummy_start(),
                                    chan.dummy_stop(),
                                    chan.dummy_n_points())
    np.random.seed(random_seed)
    expected_data = np.random.rand(n)
    if storage_type == 'array':
        expected_dep_data = expected_dep_data.reshape((1,
                                                       chan.dummy_n_points()))
        expected_data = expected_data.reshape((1, chan.dummy_n_points()))

    assert_allclose(subdata[dependency_name], expected_dep_data)
    assert_allclose(subdata['dummy_channel_inst_ChanA_'
                            'dummy_parameter_with_setpoints'],
                    expected_data)


@settings(max_examples=5, deadline=None,
          suppress_health_check=(HealthCheck.function_scoped_fixture,))
@given(n=hst.integers(min_value=5, max_value=500))
@pytest.mark.parametrize("bg_writing", [True, False])
@pytest.mark.parametrize("storage_type", ['numeric', 'array'])
@pytest.mark.usefixtures("experiment")
def test_datasaver_parameter_with_setpoints_explicitly_expanded(channel_array_instrument,
                                                     DAC, n, storage_type,
                                                     bg_writing):
    random_seed = 1
    chan = channel_array_instrument.A
    param = chan.dummy_parameter_with_setpoints
    chan.dummy_n_points(n)
    chan.dummy_start(0)
    chan.dummy_stop(100)
    meas = Measurement()
    meas.register_parameter(param, paramtype=storage_type)

    assert len(meas.parameters) == 2
    dependency_name = 'dummy_channel_inst_ChanA_dummy_sp_axis'

    dep_ps = meas.parameters[dependency_name]
    param_ps = meas.parameters[str(param)]

    assert dep_ps in meas._interdeps.dependencies[param_ps]
    assert meas.parameters[str(param)].type == storage_type
    assert meas.parameters[dependency_name].type == storage_type

    # Now for a real measurement
    with meas.run(write_in_background=bg_writing) as datasaver:
        # we seed the random number generator
        # so we can test that we get the expected numbers
        np.random.seed(random_seed)
        datasaver.add_result(*expand_setpoints_helper(param))
    if storage_type == 'numeric':
        expected_points_written = n
    elif storage_type == 'array':
        expected_points_written = 1

    assert datasaver.points_written == expected_points_written

    expected_params = (dependency_name,
                       'dummy_channel_inst_ChanA_dummy_parameter_with_setpoints')
    ds = load_by_id(datasaver.run_id)
    loaded_data = ds.get_parameter_data()
    for param in expected_params:
        data = loaded_data['dummy_channel_inst_ChanA_dummy_parameter_with_setpoints'][param]
        if storage_type == 'array':
            assert data.shape == (expected_points_written, n)
        else:
            assert data.shape == (expected_points_written,)
    assert len(loaded_data) == 1

    subdata = loaded_data[
        'dummy_channel_inst_ChanA_dummy_parameter_with_setpoints']

    expected_dep_data = np.linspace(chan.dummy_start(),
                                    chan.dummy_stop(),
                                    chan.dummy_n_points())
    np.random.seed(random_seed)
    expected_data = np.random.rand(n)
    if storage_type == 'array':
        expected_dep_data = expected_dep_data.reshape((1,
                                                       chan.dummy_n_points()))
        expected_data = expected_data.reshape((1, chan.dummy_n_points()))

    assert_allclose(subdata[dependency_name], expected_dep_data)
    assert_allclose(subdata['dummy_channel_inst_ChanA_'
                            'dummy_parameter_with_setpoints'],
                    expected_data)


@pytest.mark.usefixtures("experiment")
def test_datasaver_parameter_with_setpoints_partially_expanded_raises(channel_array_instrument, DAC):
    random_seed = 1
    chan = channel_array_instrument.A
    param = chan.dummy_parameter_with_setpoints_2d
    chan.dummy_n_points(10)
    chan.dummy_n_points_2(20)
    chan.dummy_start(0)
    chan.dummy_stop(100)
    chan.dummy_start_2(5)
    chan.dummy_stop_2(7)
    meas = Measurement()
    meas.register_parameter(param)

    sp_param_1 = chan.dummy_sp_axis

    assert len(meas.parameters) == 3
    dependency_name = 'dummy_channel_inst_ChanA_dummy_sp_axis'

    dep_ps = meas.parameters[dependency_name]
    param_ps = meas.parameters[str(param)]

    assert dep_ps in meas._interdeps.dependencies[param_ps]

    with meas.run() as datasaver:
        # we seed the random number generator
        # so we can test that we get the expected numbers
        np.random.seed(random_seed)
        with pytest.raises(ValueError, match="Some of the setpoints of"):
            datasaver.add_result((param, param.get()),
                                 (sp_param_1, sp_param_1.get()))


@pytest.mark.parametrize("bg_writing", [True, False])
@settings(max_examples=5, deadline=None,
          suppress_health_check=(HealthCheck.function_scoped_fixture,))
@given(n=hst.integers(min_value=5, max_value=500))
@pytest.mark.usefixtures("experiment")
def test_datasaver_parameter_with_setpoints_complex(channel_array_instrument,
                                                    DAC, n,
                                                    bg_writing):
    random_seed = 1
    chan = channel_array_instrument.A
    param = chan.dummy_parameter_with_setpoints_complex
    chan.dummy_n_points(n)
    chan.dummy_start(0)
    chan.dummy_stop(100)
    meas = Measurement()
    meas.register_parameter(param, paramtype='array')

    assert len(meas.parameters) == 2

    dependency_name = 'dummy_channel_inst_ChanA_dummy_sp_axis'

    dependent_parameter = meas.parameters[str(param)]
    indepdendent_parameter = meas.parameters[dependency_name]

    assert meas._interdeps.dependencies[dependent_parameter] \
           == (indepdendent_parameter, )

    assert dependent_parameter.type == 'array'
    assert indepdendent_parameter.type == 'array'

    # Now for a real measurement
    with meas.run(write_in_background=bg_writing) as datasaver:
        # we seed the random number generator
        # so we can test that we get the expected numbers
        np.random.seed(random_seed)
        datasaver.add_result((param, param.get()))
    assert datasaver.points_written == 1

    ds = load_by_id(datasaver.run_id)
    datadict = ds.get_parameter_data()
    assert len(datadict) == 1
    subdata = datadict[
        'dummy_channel_inst_ChanA_dummy_parameter_with_setpoints_complex']
    assert_allclose(subdata[dependency_name],
                    np.linspace(chan.dummy_start(),
                                chan.dummy_stop(),
                                chan.dummy_n_points()).reshape(1, chan.dummy_n_points()))
    np.random.seed(random_seed)
    assert_allclose(subdata['dummy_channel_inst_ChanA_'
                            'dummy_parameter_with_setpoints_complex'],
                    (np.random.rand(n) + 1j * np.random.rand(n)).reshape(1, chan.dummy_n_points()))


@pytest.mark.parametrize("bg_writing", [True, False])
@settings(max_examples=5, deadline=None,
          suppress_health_check=(HealthCheck.function_scoped_fixture,))
@given(n=hst.integers(min_value=5, max_value=500))
@pytest.mark.usefixtures("experiment")
def test_datasaver_parameter_with_setpoints_complex_explicitly_expanded(channel_array_instrument,
                                                             DAC, n,
                                                             bg_writing):
    random_seed = 1
    chan = channel_array_instrument.A
    param = chan.dummy_parameter_with_setpoints_complex
    chan.dummy_n_points(n)
    chan.dummy_start(0)
    chan.dummy_stop(100)
    meas = Measurement()
    meas.register_parameter(param, paramtype='array')

    assert len(meas.parameters) == 2

    dependency_name = 'dummy_channel_inst_ChanA_dummy_sp_axis'

    dependent_parameter = meas.parameters[str(param)]
    indepdendent_parameter = meas.parameters[dependency_name]

    assert meas._interdeps.dependencies[dependent_parameter] \
           == (indepdendent_parameter, )

    assert dependent_parameter.type == 'array'
    assert indepdendent_parameter.type == 'array'

    # Now for a real measurement
    with meas.run(write_in_background=bg_writing) as datasaver:
        # we seed the random number generator
        # so we can test that we get the expected numbers
        np.random.seed(random_seed)
        datasaver.add_result(*expand_setpoints_helper(param))
    assert datasaver.points_written == 1

    ds = load_by_id(datasaver.run_id)
    datadict = ds.get_parameter_data()
    assert len(datadict) == 1
    subdata = datadict[
        'dummy_channel_inst_ChanA_dummy_parameter_with_setpoints_complex']
    assert_allclose(subdata[dependency_name],
                    np.linspace(chan.dummy_start(),
                                chan.dummy_stop(),
                                chan.dummy_n_points()).reshape(1, chan.dummy_n_points()))
    np.random.seed(random_seed)
    assert_allclose(subdata['dummy_channel_inst_ChanA_'
                            'dummy_parameter_with_setpoints_complex'],
                    (np.random.rand(n) + 1j * np.random.rand(n)).reshape(1, chan.dummy_n_points()))


@pytest.mark.parametrize("bg_writing", [True, False])
@pytest.mark.parametrize("storage_type", ['numeric', 'array'])
@pytest.mark.usefixtures("experiment")
def test_datasaver_parameter_with_setpoints_missing_reg_raises(
        channel_array_instrument,
        DAC, storage_type,
        bg_writing):
    """
    Test that if for whatever reason new setpoints are added after
    registering but before adding this raises correctly
    """
    chan = channel_array_instrument.A
    param = chan.dummy_parameter_with_setpoints
    chan.dummy_n_points(11)
    chan.dummy_start(0)
    chan.dummy_stop(10)

    old_setpoints = param.setpoints
    param.setpoints = ()
    meas = Measurement()
    meas.register_parameter(param, paramtype=storage_type)

    param.setpoints = old_setpoints
    with meas.run(write_in_background=bg_writing) as datasaver:
        sp_param_name = 'dummy_channel_inst_ChanA_dummy_sp_axis'
        match = re.escape('Can not add result for parameter '
                          f'{sp_param_name}, no such parameter registered '
                          'with this measurement.')
        with pytest.raises(ValueError, match=match):
            datasaver.add_result(*expand_setpoints_helper(param))

    with meas.run(write_in_background=bg_writing) as datasaver:
        sp_param_name = 'dummy_channel_inst_ChanA_dummy_sp_axis'
        match = re.escape('Can not add result for parameter '
                          f'{sp_param_name}, no such parameter registered '
                          'with this measurement.')
        with pytest.raises(ValueError, match=match):
            datasaver.add_result((param, param.get()))


@pytest.mark.parametrize("bg_writing", [True, False])
@pytest.mark.parametrize("storage_type", ['numeric', 'array'])
@pytest.mark.usefixtures("experiment")
def test_datasaver_parameter_with_setpoints_reg_but_missing_validator(
        channel_array_instrument,
        DAC, storage_type,
        bg_writing):
    """
    Test that if for whatever reason the setpoints are removed between
    registering and adding this raises correctly. This tests tests that
    the parameter validator correctly asserts this.
    """
    chan = channel_array_instrument.A
    param = chan.dummy_parameter_with_setpoints
    chan.dummy_n_points(11)
    chan.dummy_start(0)
    chan.dummy_stop(10)

    meas = Measurement()
    meas.register_parameter(param, paramtype=storage_type)

    param.setpoints = ()

    with meas.run(write_in_background=bg_writing) as datasaver:
        with pytest.raises(ValueError, match=r"Shape of output is not"
                                             r" consistent with setpoints."
                                             r" Output is shape "
                                             r"\(<qcodes.instrument.parameter."
                                             r"Parameter: dummy_n_points at "
                                             r"[0-9]+>,\) and setpoints are "
                                             r"shape \(\)', 'getting dummy_"
                                             r"channel_inst_ChanA_dummy_"
                                             r"parameter_with_setpoints"):
            datasaver.add_result(*expand_setpoints_helper(param))

    with meas.run(write_in_background=bg_writing) as datasaver:
        with pytest.raises(ValueError, match=r"Shape of output is not"
                                             r" consistent with setpoints."
                                             r" Output is shape "
                                             r"\(<qcodes.instrument.parameter."
                                             r"Parameter: dummy_n_points at "
                                             r"[0-9]+>,\) and setpoints are "
                                             r"shape \(\)', 'getting dummy_"
                                             r"channel_inst_ChanA_dummy_"
                                             r"parameter_with_setpoints"):
            datasaver.add_result((param, param.get()))


@pytest.mark.parametrize("bg_writing", [True, False])
@pytest.mark.parametrize("storage_type", ['numeric', 'array'])
@pytest.mark.usefixtures("experiment")
def test_datasaver_parameter_with_setpoints_reg_but_missing(
        channel_array_instrument,
        DAC, storage_type,
        bg_writing):
    """
    Test that if for whatever reason the setpoints of a QCoDeS parameter are
    removed between registering that parameter with the Measurement and adding
    results for those setpoints, then then datasaver correctly raises.
    """
    chan = channel_array_instrument.A
    param = chan.dummy_parameter_with_setpoints
    chan.dummy_n_points(11)
    chan.dummy_start(0)
    chan.dummy_stop(10)

    someparam = Parameter('someparam', vals=Arrays(shape=(10,)))
    old_setpoints = param.setpoints
    param.setpoints = (old_setpoints[0], someparam)

    meas = Measurement()
    meas.register_parameter(param, paramtype=storage_type)

    param.setpoints = old_setpoints
    with meas.run(write_in_background=bg_writing) as datasaver:
        match = re.escape('Can not add result, some required parameters '
                          'are missing.')
        with pytest.raises(ValueError, match=match):
            datasaver.add_result(*expand_setpoints_helper(param))

    with meas.run(write_in_background=bg_writing) as datasaver:
        match = re.escape('Can not add result, some required parameters '
                          'are missing.')
        with pytest.raises(ValueError, match=match):
            datasaver.add_result((param, param.get()))


@settings(max_examples=5, deadline=None,
          suppress_health_check=(HealthCheck.function_scoped_fixture,))
@given(N=hst.integers(min_value=5, max_value=500))
@pytest.mark.usefixtures("experiment")
@pytest.mark.parametrize("storage_type", ['numeric', 'array'])
@pytest.mark.parametrize("bg_writing", [True, False])
def test_datasaver_array_parameters_array(channel_array_instrument, DAC, N,
                                          storage_type,
                                          bg_writing):
    """
    Test that storing array parameters inside a loop works as expected
    """
    meas = Measurement()

    array_param = channel_array_instrument.A.dummy_array_parameter

    meas.register_parameter(array_param, paramtype=storage_type)

    assert len(meas.parameters) == 2
    dependency_name = 'dummy_channel_inst_ChanA_array_setpoint_param_this_setpoint'
    dependency_ps = meas.parameters[dependency_name]
    array_param_ps = meas.parameters[str(array_param)]
    assert dependency_ps in meas._interdeps.dependencies[array_param_ps]
    assert meas.parameters[str(array_param)].type == storage_type
    assert meas.parameters[dependency_name].type == storage_type

    # Now for a real measurement

    meas = Measurement()

    meas.register_parameter(DAC.ch1, paramtype='numeric')
    meas.register_parameter(array_param, setpoints=[DAC.ch1], paramtype=storage_type)

    assert len(meas.parameters) == 3

    M = array_param.shape[0]
    dac_datapoints = np.linspace(0, 0.01, N)
    with meas.run(write_in_background=bg_writing) as datasaver:
        for set_v in dac_datapoints:
            datasaver.add_result((DAC.ch1, set_v),
                                 (array_param, array_param.get()))

    if storage_type == 'numeric':
        expected_npoints = N*M
    elif storage_type == 'array':
        expected_npoints = N

    assert datasaver.points_written == expected_npoints
    ds = load_by_id(datasaver.run_id)
    loaded_data = ds.get_parameter_data()['dummy_channel_inst_ChanA_dummy_array_parameter']

    data_num = loaded_data['dummy_dac_ch1']
    assert len(data_num) == expected_npoints

    setpoint_arrays = loaded_data[dependency_name]
    data_arrays = loaded_data['dummy_channel_inst_ChanA_dummy_array_parameter']
    assert len(setpoint_arrays) == expected_npoints
    assert len(data_arrays) == expected_npoints

    expected_dac_data = np.repeat(np.linspace(0, 0.01, N), M)
    expected_sp_data = np.tile(array_param.setpoints[0], N)
    expected_output = np.array([array_param.get() for _ in range(N)]).reshape(
        N * M)

    if storage_type == 'array':
        expected_dac_data = expected_dac_data.reshape(N, M)
        expected_sp_data = expected_sp_data.reshape(N, M)
        expected_output = expected_output.reshape(N, M)

    assert_allclose(loaded_data['dummy_dac_ch1'], expected_dac_data)
    assert_allclose(loaded_data[dependency_name],
                    expected_sp_data)
    assert_allclose(loaded_data['dummy_channel_inst_ChanA_dummy_array_parameter'],
                    expected_output)

    if storage_type == 'array':
        for data_array, setpoint_array in zip(data_arrays, setpoint_arrays):
            assert_array_equal(setpoint_array, np.linspace(5, 9, 5))
            assert_array_equal(data_array, np.array([2., 2., 2., 2., 2.]))


@pytest.mark.parametrize("bg_writing", [True, False])
@settings(max_examples=5, deadline=None,
          suppress_health_check=(HealthCheck.function_scoped_fixture,))
@given(N=hst.integers(min_value=5, max_value=500))
@pytest.mark.usefixtures("experiment")
def test_datasaver_complex_array_parameters_array(channel_array_instrument,
                                                  DAC, N,
                                                  bg_writing):
    """
    Test that storing a complex array parameters inside a loop with the sqlite
    Array type works as expected
    """
    meas = Measurement()

    array_param = channel_array_instrument.A.dummy_complex_array_parameter

    meas.register_parameter(array_param, paramtype='array')

    assert len(meas.parameters) == 2
    dependency_name = 'dummy_channel_inst_ChanA_this_setpoint'

    dependent_parameter = meas.parameters[str(array_param)]
    indepdendent_parameter = meas.parameters[dependency_name]

    assert meas._interdeps.dependencies[dependent_parameter] \
           == (indepdendent_parameter, )

    assert dependent_parameter.type == 'array'
    assert indepdendent_parameter.type == 'array'

    # Now for a real measurement

    meas = Measurement()

    meas.register_parameter(DAC.ch1, paramtype='numeric')
    meas.register_parameter(array_param, setpoints=[DAC.ch1], paramtype='array')

    assert len(meas.parameters) == 3

    M = array_param.shape[0]
    dac_datapoints = np.linspace(0, 0.01, N)
    with meas.run(write_in_background=bg_writing) as datasaver:
        for set_v in dac_datapoints:
            datasaver.add_result((DAC.ch1, set_v),
                                 (array_param, array_param.get()))
    assert datasaver.points_written == N
    ds = load_by_id(datasaver.run_id)
    loaded_data = ds.get_parameter_data()["dummy_channel_inst_ChanA_dummy_complex_array_parameter"]
    data_num = loaded_data['dummy_dac_ch1']
    assert data_num.shape == (N, M)

    param_name = 'dummy_channel_inst_ChanA_dummy_complex_array_parameter'

    setpoint_arrays = loaded_data['dummy_channel_inst_ChanA_this_setpoint']
    data_arrays = loaded_data[param_name]
    assert setpoint_arrays.shape == (N, M)
    assert data_arrays.shape == (N, M)

    for data_array, setpoint_array in zip(data_arrays, setpoint_arrays):
        assert_array_equal(setpoint_array, np.linspace(5, 9, 5))
        assert_array_equal(data_array, np.arange(5) - 1j*np.arange(5))


@pytest.mark.parametrize("bg_writing", [True, False])
def test_datasaver_multidim_array(experiment, bg_writing):  # noqa: F811
    """
    Test that inserting multidim parameters as arrays works as expected
    """
    meas = Measurement(experiment)
    size1 = 10
    size2 = 15

    data_mapping = {name: i for i, name in
                    zip(range(4), ['x1', 'x2', 'y1', 'y2'])}

    x1 = qc.ManualParameter('x1')
    x2 = qc.ManualParameter('x2')
    y1 = qc.ManualParameter('y1')
    y2 = qc.ManualParameter('y2')

    meas.register_parameter(x1, paramtype='array')
    meas.register_parameter(x2, paramtype='array')
    meas.register_parameter(y1, setpoints=[x1, x2], paramtype='array')
    meas.register_parameter(y2, setpoints=[x1, x2], paramtype='array')
    data = np.random.rand(4, size1, size2)
    expected = {'x1': data[0, :, :],
                'x2': data[1, :, :],
                'y1': data[2, :, :],
                'y2': data[3, :, :]}
    with meas.run(write_in_background=bg_writing) as datasaver:
        datasaver.add_result((str(x1), expected['x1']),
                             (str(x2), expected['x2']),
                             (str(y1), expected['y1']),
                             (str(y2), expected['y2']))

    # We expect one "point" i.e. row in the DB to be written per top-level
    # parameter.
    assert datasaver.points_written == 2
    dataset = load_by_id(datasaver.run_id)
    loaded_data = dataset.get_parameter_data()
    for outerid in ('y1', 'y2'):
        for innerid in ('x1', 'x2', outerid):
            mydata = loaded_data[outerid][innerid]
            assert mydata.shape == (1, size1, size2)
            assert_array_equal(mydata[0], expected[innerid])


@pytest.mark.parametrize("bg_writing", [True, False])
@pytest.mark.parametrize("export", [True, False])
def test_datasaver_export(experiment, bg_writing, tmp_path_factory,
                          export):
    """
    Test export data to csv after measurement ends
    """
    meas = Measurement(experiment)
    size1 = 10
    size2 = 15

    x1 = qc.ManualParameter('x1')
    x2 = qc.ManualParameter('x2')
    y1 = qc.ManualParameter('y1')
    y2 = qc.ManualParameter('y2')

    meas.register_parameter(x1, paramtype='array')
    meas.register_parameter(x2, paramtype='array')
    meas.register_parameter(y1, setpoints=[x1, x2], paramtype='array')
    meas.register_parameter(y2, setpoints=[x1, x2], paramtype='array')
    data = np.random.rand(4, size1, size2)
    expected = {'x1': data[0, :, :],
                'x2': data[1, :, :],
                'y1': data[2, :, :],
                'y2': data[3, :, :]}

    tmp_path = tmp_path_factory.mktemp("export_from_config")
    path = str(tmp_path)

    with patch("qcodes.dataset.data_set.get_data_export_type") as mock_type, \
    patch("qcodes.dataset.data_set.get_data_export_path") as mock_path, \
    patch("qcodes.dataset.measurements.get_data_export_automatic") as mock_automatic:
        mock_type.return_value = DataExportType.CSV
        mock_path.return_value = path
        mock_automatic.return_value = export
        with meas.run(write_in_background=bg_writing) as datasaver:
            datasaver.add_result((str(x1), expected['x1']),
                                 (str(x2), expected['x2']),
                                 (str(y1), expected['y1']),
                                 (str(y2), expected['y2']))
    if export:
        assert os.listdir(path) == [f"qcodes_{datasaver.dataset.run_id}.csv"]
    else:
        assert os.listdir(path) == []


@pytest.mark.parametrize("bg_writing", [True, False])
def test_datasaver_multidim_numeric(experiment, bg_writing):
    """
    Test that inserting multidim parameters as numeric works as expected
    """
    meas = Measurement(experiment)
    size1 = 10
    size2 = 15
    x1 = qc.ManualParameter('x1')
    x2 = qc.ManualParameter('x2')
    y1 = qc.ManualParameter('y1')
    y2 = qc.ManualParameter('y2')

    data_mapping = {name: i for i, name in
                    zip(range(4), ['x1', 'x2', 'y1', 'y2'])}

    meas.register_parameter(x1, paramtype='numeric')
    meas.register_parameter(x2, paramtype='numeric')
    meas.register_parameter(y1, setpoints=[x1, x2], paramtype='numeric')
    meas.register_parameter(y2, setpoints=[x1, x2], paramtype='numeric')
    data = np.random.rand(4, size1, size2)
    with meas.run(write_in_background=bg_writing) as datasaver:
        datasaver.add_result((str(x1), data[0, :, :]),
                             (str(x2), data[1, :, :]),
                             (str(y1), data[2, :, :]),
                             (str(y2), data[3, :, :]))
    # The factor of 2 is due to there being 2 top-level params
    assert datasaver.points_written == 2 * (size1 * size2)
    dataset = load_by_id(datasaver.run_id)
    all_data = dataset.get_parameter_data()
    for outer in ('y1', 'y2'):
        for inner in ('x1', 'x2', outer):
            mydata = all_data[outer][inner]
            assert mydata.shape == (size1 * size2, )
            assert mydata.dtype == np.float64


@pytest.mark.parametrize("bg_writing", [True, False])
@pytest.mark.usefixtures("experiment")
def test_datasaver_multidimarrayparameter_as_array(SpectrumAnalyzer,
                                                   bg_writing):
    """
    Test that inserting multidim Arrrayparameters as array works as expected
    """
    array_param = SpectrumAnalyzer.multidimspectrum
    meas = Measurement()
    meas.register_parameter(array_param, paramtype='array')
    assert len(meas.parameters) == 4
    inserted_data = array_param.get()
    with meas.run(write_in_background=bg_writing) as datasaver:
        datasaver.add_result((array_param, inserted_data))

    assert datasaver.points_written == 1
    ds = load_by_id(datasaver.run_id)
    expected_shape = (1, 100, 50, 20)
    loaded_data = ds.get_parameter_data()
    for i in range(3):
        data = loaded_data['dummy_SA_multidimspectrum'][f'dummy_SA_Frequency{i}']
        aux_shape = list(expected_shape)
        aux_shape.pop(i+1)

        assert data.shape == expected_shape
        for j in range(aux_shape[1]):
            for k in range(aux_shape[2]):
                # todo There should be a simpler way of doing this
                if i == 0:
                    mydata = data[0, :, j, k]
                if i == 1:
                    mydata = data[0, j, :, k]
                if i == 2:
                    mydata = data[0, j, k, :]
                assert_array_equal(mydata,
                                   np.linspace(array_param.start,
                                               array_param.stop,
                                               array_param.npts[i]))


@pytest.mark.parametrize("bg_writing", [True, False])
@pytest.mark.usefixtures("experiment")
def test_datasaver_multidimarrayparameter_as_numeric(SpectrumAnalyzer,
                                                     bg_writing):
    """
    Test that storing a multidim Array parameter as numeric unravels the
    parameter as expected.
    """

    array_param = SpectrumAnalyzer.multidimspectrum
    meas = Measurement()
    meas.register_parameter(array_param, paramtype='numeric')
    expected_shape = array_param.shape
    dims = len(array_param.shape)
    assert len(meas.parameters) == dims + 1

    points_expected = np.prod(array_param.npts)
    inserted_data = array_param.get()
    with meas.run(write_in_background=bg_writing) as datasaver:
        datasaver.add_result((array_param, inserted_data))

    assert datasaver.points_written == points_expected
    ds = load_by_id(datasaver.run_id)
    # check setpoints

    expected_setpoints_vectors = (np.linspace(array_param.start,
                                              array_param.stop,
                                              array_param.npts[i]) for i in
                                  range(dims))
    expected_setpoints_matrix = np.meshgrid(*expected_setpoints_vectors,
                                            indexing='ij')
    expected_setpoints = tuple(
        setpoint_array.ravel() for setpoint_array in expected_setpoints_matrix)

    loaded_data = ds.get_parameter_data()

    for i in range(dims):
        data = loaded_data['dummy_SA_multidimspectrum'][f'dummy_SA_Frequency{i}']
        assert len(data) == points_expected
        assert_allclose(data.squeeze(),
                        expected_setpoints[i])
    data = loaded_data['dummy_SA_multidimspectrum']['dummy_SA_multidimspectrum'].squeeze()
    assert_allclose(data, inserted_data.ravel())


@pytest.mark.parametrize("bg_writing", [True, False])
@pytest.mark.usefixtures("experiment")
def test_datasaver_multi_parameters_scalar(channel_array_instrument,
                                           bg_writing):
    """
    Test that we can register multiparameters that are scalar.
    """
    meas = Measurement()
    param = channel_array_instrument.A.dummy_scalar_multi_parameter
    meas.register_parameter(param)
    assert len(meas.parameters) == len(param.shapes)
    assert set(meas.parameters.keys()) == set(param.full_names)

    with meas.run(write_in_background=bg_writing) as datasaver:
        datasaver.add_result((param, param()))

    assert datasaver.points_written == 2
    ds = load_by_id(datasaver.run_id)
    assert ds.get_parameter_data()['dummy_channel_inst_ChanA_thisparam']['dummy_channel_inst_ChanA_thisparam'] == np.array([[0]])
    assert ds.get_parameter_data()['dummy_channel_inst_ChanA_thatparam']['dummy_channel_inst_ChanA_thatparam'] == np.array([[1]])


@pytest.mark.parametrize("bg_writing", [True, False])
@pytest.mark.usefixtures("experiment")
def test_datasaver_multi_parameters_array(channel_array_instrument,
                                          bg_writing):
    """
    Test that we can register multiparameters that are array like.
    """
    meas = Measurement()
    param = channel_array_instrument.A.dummy_multi_parameter
    meas.register_parameter(param)
    assert len(meas.parameters) == 3  # two params + 1D identical setpoints
    param_names = ('dummy_channel_inst_ChanA_multi_setpoint_param_this_setpoint',
                   'dummy_channel_inst_ChanA_multi_setpoint_param_this',
                   'dummy_channel_inst_ChanA_multi_setpoint_param_that')
    assert set(meas.parameters.keys()) == set(param_names)
    this_ps = meas.parameters[param_names[1]]
    that_ps = meas.parameters[param_names[2]]
    sp_ps = meas.parameters[param_names[0]]
    assert sp_ps in meas._interdeps.dependencies[this_ps]
    assert sp_ps in meas._interdeps.dependencies[that_ps]

    with meas.run(write_in_background=bg_writing) as datasaver:
        datasaver.add_result((param, param()))
    assert datasaver.points_written == 2 * 5
    ds = load_by_id(datasaver.run_id)
    setpts = np.arange(5, 10)

    np.testing.assert_array_equal(ds.get_parameter_data()[param_names[1]][param_names[0]], setpts)
    np.testing.assert_array_equal(ds.get_parameter_data()[param_names[2]][param_names[0]], setpts)

    this_read_data = ds.get_parameter_data()[param_names[1]][param_names[1]]
    that_read_data = ds.get_parameter_data()[param_names[2]][param_names[2]]
    np.testing.assert_array_equal(this_read_data, np.zeros(5))
    np.testing.assert_array_equal(that_read_data, np.ones(5))


@pytest.mark.parametrize("bg_writing", [True, False])
@pytest.mark.usefixtures("experiment")
def test_datasaver_2d_multi_parameters_array(channel_array_instrument,
                                             bg_writing):
    """
    Test that we can register multiparameters that are array like and 2D.
    """
    sp_name_1 = "dummy_channel_inst_ChanA_multi_2d_setpoint_param_this_setpoint"
    sp_name_2 = 'dummy_channel_inst_ChanA_multi_2d_setpoint_param_that_setpoint'
    p_name_1 = 'dummy_channel_inst_ChanA_this'
    p_name_2 = 'dummy_channel_inst_ChanA_that'
    from functools import reduce

    meas = Measurement()
    param = channel_array_instrument.A.dummy_2d_multi_parameter
    meas.register_parameter(param)
    assert len(meas.parameters) == 4  # two params + 2D identical setpoints
    param_names = (sp_name_1,
                   sp_name_2,
                   p_name_1,
                   p_name_2)
    assert set(meas.parameters.keys()) == set(param_names)
    this_ps = meas.parameters[p_name_1]
    that_ps = meas.parameters[p_name_2]
    this_sp_ps = meas.parameters[sp_name_1]
    that_sp_ps = meas.parameters[sp_name_2]
    assert that_sp_ps in meas._interdeps.dependencies[this_ps]
    assert that_sp_ps in meas._interdeps.dependencies[that_ps]
    assert this_sp_ps in meas._interdeps.dependencies[this_ps]
    assert this_sp_ps in meas._interdeps.dependencies[that_ps]

    with meas.run(write_in_background=bg_writing) as datasaver:
        datasaver.add_result((param, param()))

    assert datasaver.points_written == 2 * 15
    ds = load_by_id(datasaver.run_id)

    # 30 points in each setpoint value list
    this_sp_val = np.array(reduce(list.__add__, [[n]*3 for n in range(5, 10)], []))
    that_sp_val = np.array(reduce(list.__add__, [[n] for n in range(9, 12)], []) * 5)

    np.testing.assert_array_equal(
        ds.get_parameter_data()[p_name_1][sp_name_1],
        this_sp_val
    )
    np.testing.assert_array_equal(
        ds.get_parameter_data()[p_name_1][sp_name_2],
        that_sp_val
    )
    np.testing.assert_array_equal(
        ds.get_parameter_data()[p_name_2][sp_name_1],
        this_sp_val
    )
    np.testing.assert_array_equal(
        ds.get_parameter_data()[p_name_2][sp_name_2],
        that_sp_val
    )

    this_read_data = ds.get_parameter_data()[p_name_1][p_name_1]
    that_read_data = ds.get_parameter_data()[p_name_2][p_name_2]
    assert len(this_read_data) == 15
    assert len(that_read_data) == 15

    np.testing.assert_array_equal(this_read_data, np.zeros(15))
    np.testing.assert_array_equal(that_read_data, np.ones(15))


@pytest.mark.usefixtures("experiment")
@pytest.mark.parametrize("bg_writing", [True, False])
@pytest.mark.parametrize("storage_type", ['numeric', 'array'])
@settings(deadline=None)
@given(Ns=hst.lists(hst.integers(2, 10), min_size=2, max_size=5))
def test_datasaver_arrays_of_different_length(storage_type, Ns, bg_writing):
    """
    Test that we can save arrays of different length in a single call to
    datasaver.add_result
    """

    no_of_signals = len(Ns)

    meas = Measurement()
    meas.register_custom_parameter('temperature',
                                   paramtype='numeric',
                                   label='Temperature',
                                   unit='K')
    for n in range(no_of_signals):
        meas.register_custom_parameter(f'freqs{n}', paramtype=storage_type)
        meas.register_custom_parameter(f'signal{n}',
                                       paramtype=storage_type,
                                       setpoints=(f'freqs{n}', 'temperature'))

    with meas.run(write_in_background=bg_writing) as datasaver:
        result_t = ('temperature', 70)
        result_freqs = list((f'freqs{n}', np.linspace(0, 1, Ns[n]))
                              for n in range(no_of_signals))
        result_sigs = list((f'signal{n}', np.random.randn(Ns[n]))
                             for n in range(no_of_signals))
        full_result = tuple(result_freqs + result_sigs + [result_t])
        datasaver.add_result(*full_result)

    ds = load_by_id(datasaver.run_id)

    data = ds.get_parameter_data()

    assert list(data.keys()) == [f'signal{n}' for n in range(no_of_signals)]
    for n in range(no_of_signals):
        assert (data[f'signal{n}']['temperature'] == np.array([70]*(Ns[n]))).all()


@pytest.mark.parametrize("bg_writing", [True, False])
@pytest.mark.usefixtures("experiment")
def test_save_complex_num(complex_num_instrument, bg_writing):
    """
    Test that we can save various parameters mixed with complex parameters
    """

    # scalar complex parameter
    setparam = complex_num_instrument.setpoint
    param = complex_num_instrument.complex_num
    # array parameter
    arrayparam = complex_num_instrument.some_array
    # complex array parameter
    complexarrayparam = complex_num_instrument.some_complex_array
    some_complex_array_setpoints = complex_num_instrument.some_complex_array_setpoints

    meas = Measurement()
    meas.register_parameter(setparam, paramtype='numeric')
    meas.register_parameter(param, paramtype='complex', setpoints=(setparam,))
    meas.register_parameter(arrayparam, paramtype='array',
                            setpoints=(setparam,))

    meas.register_parameter(some_complex_array_setpoints, paramtype='numeric')
    meas.register_parameter(complexarrayparam, paramtype='complex',
                            setpoints=(setparam, some_complex_array_setpoints))

    with meas.run(write_in_background=bg_writing) as datasaver:
        for i in range(10):
            setparam.set(i)
            datasaver.add_result((setparam, setparam()),
                                 (param, param()),
                                 *expand_setpoints_helper(arrayparam),
                                 (some_complex_array_setpoints, some_complex_array_setpoints.get()),
                                 (complexarrayparam, complexarrayparam.get()))

    data = datasaver.dataset.get_parameter_data()

    # scalar complex parameter
    setpoints_num = data['dummy_channel_inst_complex_num'][
        'dummy_channel_inst_setpoint']
    data_num = data['dummy_channel_inst_complex_num'][
        'dummy_channel_inst_complex_num']

    assert_allclose(setpoints_num, np.arange(10))
    assert_allclose(data_num, np.arange(10) + 1j*np.arange(10))

    # array parameter
    setpoints1_array = data['dummy_channel_inst_some_array'][
        'dummy_channel_inst_setpoint']
    assert_allclose(setpoints1_array, np.repeat(np.arange(10), 5).reshape(10, 5))

    setpoints2_array = data['dummy_channel_inst_some_array'][
        'dummy_channel_inst_some_array_setpoints']

    assert_allclose(setpoints2_array, np.tile(np.arange(5), 10).reshape(10, 5))

    array_data = data['dummy_channel_inst_some_array'][
        'dummy_channel_inst_some_array']

    assert_allclose(array_data, np.ones((10, 5)))

    # complex array parameter
    setpoints1_array = data['dummy_channel_inst_some_complex_array'][
        'dummy_channel_inst_setpoint']
    assert_allclose(setpoints1_array, np.repeat(np.arange(10), 5))

    setpoints2_array = data['dummy_channel_inst_some_complex_array'][
        'dummy_channel_inst_some_complex_array_setpoints']

    assert_allclose(setpoints2_array, np.tile(np.arange(5), 10))

    array_data = data['dummy_channel_inst_some_complex_array'][
        'dummy_channel_inst_some_complex_array']

    assert_allclose(array_data, np.ones(50)+1j*np.ones(50))


@pytest.mark.parametrize("bg_writing", [True, False])
@pytest.mark.usefixtures("experiment")
def test_save_and_reload_complex_standalone(complex_num_instrument,
                                            bg_writing):
    param = complex_num_instrument.complex_num
    complex_num_instrument.setpoint(1)
    p = qc.instrument.parameter.Parameter(
        'test',
        set_cmd=None,
        get_cmd=lambda: 1+1j,
        vals=qc.utils.validators.ComplexNumbers())
    meas = qc.dataset.measurements.Measurement()
    meas.register_parameter(param)
    pval = param.get()
    with meas.run(write_in_background=bg_writing) as datasaver:
        datasaver.add_result((param, pval))
    data = datasaver.dataset.get_parameter_data()
    data_num = data['dummy_channel_inst_complex_num'][
        'dummy_channel_inst_complex_num']
    assert_allclose(data_num, 1 + 1j)


@pytest.mark.parametrize("bg_writing", [True, False])
@pytest.mark.usefixtures("experiment")
def test_save_complex_num_setpoints(complex_num_instrument, bg_writing):
    """
    Test that we can save a parameter with complex setpoints
    """
    setparam = complex_num_instrument.complex_setpoint
    param = complex_num_instrument.real_part
    meas = Measurement()
    meas.register_parameter(setparam, paramtype='complex')
    meas.register_parameter(param, paramtype='numeric', setpoints=(setparam,))

    with meas.run(write_in_background=bg_writing) as datasaver:
        for i in range(10):
            setparam.set(i+1j*i)
            datasaver.add_result((setparam, setparam()),
                                 (param, param()))
    data = datasaver.dataset.get_parameter_data()
    setpoints_num = data['dummy_channel_inst_real_part'][
        'dummy_channel_inst_complex_setpoint']
    data_num = data['dummy_channel_inst_real_part'][
        'dummy_channel_inst_real_part']

    assert_allclose(setpoints_num, np.arange(10) + 1j*np.arange(10))
    assert_allclose(data_num, np.arange(10))


@pytest.mark.parametrize("bg_writing", [True, False])
@pytest.mark.usefixtures("experiment")
def test_save_complex_num_setpoints_array(complex_num_instrument, bg_writing):
    """
    Test that we can save an array parameter with complex setpoints
    """

    setparam = complex_num_instrument.complex_setpoint
    param = complex_num_instrument.some_array

    meas = Measurement()
    meas.register_parameter(setparam, paramtype='complex')
    meas.register_parameter(param, paramtype='array', setpoints=(setparam,))

    with meas.run(write_in_background=bg_writing) as datasaver:
        for i in range(10):
            setparam.set(i+1j*i)
            datasaver.add_result((setparam, setparam()),
                                 *expand_setpoints_helper(param))
    data = datasaver.dataset.get_parameter_data()
    setpoints1 = data['dummy_channel_inst_some_array'][
        'dummy_channel_inst_complex_setpoint']
    setpoints2 = data['dummy_channel_inst_some_array'][
        'dummy_channel_inst_some_array_setpoints']
    data_num = data['dummy_channel_inst_some_array'][
        'dummy_channel_inst_some_array']

    assert_allclose(setpoints1, np.repeat(np.arange(10) +
                                          1j*np.arange(10), 5).reshape((10, 5)))
    assert_allclose(setpoints2, np.tile(np.arange(5), 10).reshape((10, 5)))

    assert_allclose(data_num, np.ones((10, 5)))


@pytest.mark.parametrize("bg_writing", [True, False])
@pytest.mark.usefixtures("experiment")
def test_save_complex_as_num_raises(complex_num_instrument, bg_writing):
    setparam = complex_num_instrument.setpoint
    param = complex_num_instrument.complex_num
    meas = Measurement()
    meas.register_parameter(setparam, paramtype='numeric')
    meas.register_parameter(param, paramtype='numeric', setpoints=(setparam,))

    expected_msg = ('Parameter dummy_channel_inst_complex_num is of '
                    'type "numeric", but got a result of '
                    'type complex128')

    with meas.run(write_in_background=bg_writing) as datasaver:
        setparam.set(0)
        with pytest.raises(ValueError, match=expected_msg):
            datasaver.add_result((setparam, setparam()),
                                 (param, param()))


@pytest.mark.parametrize("bg_writing", [True, False])
@pytest.mark.usefixtures("experiment")
def test_save_numeric_as_complex_raises(complex_num_instrument, bg_writing):
    setparam = complex_num_instrument.setpoint
    param = complex_num_instrument.complex_num
    meas = Measurement()
    meas.register_parameter(setparam, paramtype='numeric')
    meas.register_parameter(param, paramtype='complex', setpoints=(setparam,))

    expected_msg = ('Parameter dummy_channel_inst_complex_num is of '
                    'type "complex", but got a result of type int')

    with meas.run(write_in_background=bg_writing) as datasaver:
        setparam.set(0)
        with pytest.raises(ValueError, match=expected_msg):
            datasaver.add_result((setparam, setparam()),
                                 (param, setparam()))


def test_parameter_inference(channel_array_instrument):
    chan = channel_array_instrument.channels[0]
    # default values
    assert Measurement._infer_paramtype(chan.temperature, None) is None
    assert Measurement._infer_paramtype(chan.dummy_array_parameter,
                                        None) == 'array'
    assert Measurement._infer_paramtype(chan.dummy_parameter_with_setpoints,
                                        None) == 'array'
    assert Measurement._infer_paramtype(chan.dummy_multi_parameter,
                                        None) is None
    assert Measurement._infer_paramtype(chan.dummy_scalar_multi_parameter,
                                        None) is None
    assert Measurement._infer_paramtype(chan.dummy_2d_multi_parameter,
                                        None) is None
    assert Measurement._infer_paramtype(chan.dummy_text,
                                        None) == 'text'
    assert Measurement._infer_paramtype(chan.dummy_complex,
                                        None) == 'complex'

    # overwrite the default with sensible alternatives
    assert Measurement._infer_paramtype(chan.dummy_array_parameter,
                                        'numeric') == 'numeric'
    assert Measurement._infer_paramtype(chan.dummy_parameter_with_setpoints,
                                        'numeric') == 'numeric'
    assert Measurement._infer_paramtype(chan.dummy_multi_parameter,
                                        'array') == 'array'
    assert Measurement._infer_paramtype(chan.dummy_2d_multi_parameter,
                                        'array') == 'array'


@pytest.mark.usefixtures("experiment")
def test_load_legacy_files_2D():
    location = '../fixtures/2018-01-17/#002_2D_test_15-43-14'
    dir = os.path.dirname(__file__)
    full_location = os.path.join(dir, location)
    run_ids = import_dat_file(full_location)
    run_id = run_ids[0]
    data = load_by_id(run_id)
    assert data.parameters == 'dac_ch1_set,dac_ch2_set,dmm_voltage'
    assert data.number_of_results == 36
    expected_names = ['dac_ch1_set', 'dac_ch2_set', 'dmm_voltage']
    expected_labels = ['Gate ch1', 'Gate ch2', 'Gate voltage']
    expected_units = ['V', 'V', 'V']
    expected_depends_on = ['', '', 'dac_ch1_set, dac_ch2_set']
    for i, parameter in enumerate(data.get_parameters()):
        assert parameter.name == expected_names[i]
        assert parameter.label == expected_labels[i]
        assert parameter.unit == expected_units[i]
        assert parameter.depends_on == expected_depends_on[i]
        assert parameter.type == 'numeric'
    snapshot = json.loads(data.get_metadata('snapshot'))
    assert sorted(list(snapshot.keys())) == ['__class__', 'arrays',
                                             'formatter', 'io', 'location',
                                             'loop', 'station']


@pytest.mark.usefixtures("experiment")
def test_load_legacy_files_1D():
    location = '../fixtures/2018-01-17/#001_testsweep_15-42-57'
    dir = os.path.dirname(__file__)
    full_location = os.path.join(dir, location)
    run_ids = import_dat_file(full_location)
    run_id = run_ids[0]
    data = load_by_id(run_id)
    assert data.parameters == 'dac_ch1_set,dmm_voltage'
    assert data.number_of_results == 201
    expected_names = ['dac_ch1_set', 'dmm_voltage']
    expected_labels = ['Gate ch1', 'Gate voltage']
    expected_units = ['V', 'V']
    expected_depends_on = ['', 'dac_ch1_set']
    for i, parameter in enumerate(data.get_parameters()):
        assert parameter.name == expected_names[i]
        assert parameter.label == expected_labels[i]
        assert parameter.unit == expected_units[i]
        assert parameter.depends_on == expected_depends_on[i]
        assert parameter.type == 'numeric'
    snapshot = json.loads(data.get_metadata('snapshot'))
    assert sorted(list(snapshot.keys())) == ['__class__', 'arrays',
                                             'formatter', 'io', 'location',
                                             'loop', 'station']


@pytest.mark.parametrize("bg_writing", [True, False])
@pytest.mark.usefixtures("experiment")
def test_adding_parents(bg_writing, DAC):
    """
    Test that we can register a DataSet as the parent of another DataSet
    as created by the Measurement
    """

    # The narrative of the test is that we do a measurement once, then learn
    # from the result of that where to measure next. We want to annotate the
    # second run as having the first run as predecessor


    meas = (Measurement()
            .register_parameter(DAC.ch1)
            .register_parameter(DAC.ch2, setpoints=[DAC.ch1]))

    with meas.run(write_in_background=bg_writing) as datasaver:
        datasaver.add_result((DAC.ch1, 0), (DAC.ch2, 1))

    parent_ds = datasaver.dataset

    meas = (Measurement()
            .register_parameter(DAC.ch1)
            .register_parameter(DAC.ch2, setpoints=[DAC.ch1])
            .register_parent(parent=parent_ds, link_type="predecessor"))

    with meas.run(write_in_background=bg_writing) as datasaver:
        datasaver.add_result((DAC.ch1, 1), (DAC.ch2, 2))

    child_ds = datasaver.dataset

    ds_links = child_ds.parent_dataset_links

    assert len(ds_links) == 1
    assert ds_links[0].tail == parent_ds.guid
    assert ds_links[0].head == child_ds.guid
    assert ds_links[0].edge_type == "predecessor"



import time
from datetime import datetime, timedelta

import pytest

from qcodes.instrument.parameter import Parameter, _BaseParameter
import qcodes.utils.validators as vals
from .conftest import NOT_PASSED, BetterGettableParam, SettableParam


def test_get_from_cache_does_not_trigger_real_get_if_get_if_invalid_false():
    """
    assert that calling get on the cache with get_if_invalid=False does
    not trigger a get of the parameter when parameter
    has expired due to max_val_age
    """
    param = BetterGettableParam(name="param", max_val_age=1)
    param.get()
    assert param._get_count == 1
    # let the cache expire
    time.sleep(2)
    param.cache.get(get_if_invalid=False)
    assert param._get_count == 1


def test_initial_set_with_without_cache():
    value = 43
    # setting the initial value triggers a set
    param1 = SettableParam(name="param", initial_value=value)
    assert param1._set_count == 1
    assert param1.cache.get(get_if_invalid=False) == value
    # setting the cache does not trigger a set
    param2 = SettableParam(name="param", initial_cache_value=value)
    assert param2._set_count == 0
    assert param2.cache.get(get_if_invalid=False) == value


def test_set_initial_and_initial_cache_raises():
    with pytest.raises(SyntaxError,
                       match="`initial_value` and `initial_cache_value`"):
        Parameter(name="param", initial_value=1, initial_cache_value=2)


def test_get_cache():
    time_resolution = time.get_clock_info('time').resolution
    sleep_delta = 2 * time_resolution

    # Create a gettable parameter
    local_parameter = Parameter('test_param', set_cmd=None, get_cmd=None)
    before_set = datetime.now()
    time.sleep(sleep_delta)
    local_parameter.set(1)
    time.sleep(sleep_delta)
    after_set = datetime.now()

    # Check we return last set value, with the correct timestamp
    assert local_parameter.cache.get() == 1
    assert before_set < local_parameter.cache.timestamp < after_set

    # Check that updating the value updates the timestamp
    time.sleep(sleep_delta)
    local_parameter.set(2)
    assert local_parameter.cache.get() == 2
    assert local_parameter.cache.timestamp > after_set


def test_get_cache_raw_value():
    # To have a simple distinction between raw value and value of the
    # parameter lets create a parameter with an offset
    p = Parameter('p', set_cmd=None, get_cmd=None, offset=42)
    assert p.cache.timestamp is None

    # Initially, the parameter's raw value is None
    assert p.cache.raw_value is None

    # After setting the parameter to some value, the
    # raw_value attribute of the cache should return the raw_value
    p(3)
    assert p.cache.timestamp is not None
    assert p.cache.get() == 3
    assert p.cache.raw_value == 3 + 42


def test_get_cache_unknown():
    """
    Test that cache get on a parameter that has not been acquired will
    trigger a get
    """
    value = 1
    local_parameter = BetterGettableParam('test_param', set_cmd=None,
                                          get_cmd=None)
    # fake a parameter that has a value but never been get/set to mock
    # an instrument.
    local_parameter.cache._value = value
    local_parameter.cache._raw_value = value
    assert local_parameter.cache.timestamp is None
    before_get = datetime.now()
    assert local_parameter._get_count == 0
    assert local_parameter.cache.get() == value
    assert local_parameter._get_count == 1
    # calling get above will call get since TS is None
    # and the TS will therefore no longer be None
    assert local_parameter.cache.timestamp is not None
    assert local_parameter.cache.timestamp >= before_get
    # calling cache.get now will not trigger get
    assert local_parameter.cache.get() == value
    assert local_parameter._get_count == 1


def test_get_cache_known():
    """
    Test that cache.get on a parameter that has a known value will not
    trigger a get
    """
    value = 1
    local_parameter = BetterGettableParam('test_param', set_cmd=None,
                                          get_cmd=None)
    # fake a parameter that has a value acquired 10 sec ago
    start = datetime.now()
    set_time = start - timedelta(seconds=10)
    local_parameter.cache._update_with(
        value=value, raw_value=value, timestamp=set_time)
    assert local_parameter._get_count == 0
    assert local_parameter.cache.timestamp == set_time
    assert local_parameter.cache.get() == value
    # calling cache.get above will not call get since TS is set and
    # max_val_age is not
    assert local_parameter._get_count == 0
    assert local_parameter.cache.timestamp == set_time


def test_get_cache_no_get():
    """
    Test that cache.get on a parameter that does not have get is handled
    correctly.
    """
    local_parameter = Parameter('test_param', set_cmd=None, get_cmd=False)
    # The parameter does not have a get method.
    with pytest.raises(AttributeError):
        local_parameter.get()
    # get_latest will fail as get cannot be called and no cache
    # is available
    with pytest.raises(RuntimeError):
        local_parameter.cache.get()
    value = 1
    local_parameter.set(value)
    assert local_parameter.cache.get() == value

    local_parameter2 = Parameter('test_param2', set_cmd=None,
                                 get_cmd=False, initial_value=value)
    with pytest.raises(AttributeError):
        local_parameter2.get()
    assert local_parameter2.cache.get() == value


def test_set_raw_value_on_cache():
    value = 1
    scale = 10
    local_parameter = BetterGettableParam('test_param',
                                          set_cmd=None,
                                          scale=scale)
    before = datetime.now()
    local_parameter.cache._set_from_raw_value(value*scale)
    after = datetime.now()
    assert local_parameter.cache.get(get_if_invalid=False) == value
    assert local_parameter.cache.raw_value == value * scale
    assert local_parameter.cache.timestamp >= before
    assert local_parameter.cache.timestamp <= after


def test_max_val_age():
    value = 1
    start = datetime.now()
    local_parameter = BetterGettableParam('test_param',
                                          set_cmd=None,
                                          max_val_age=1,
                                          initial_value=value)
    assert local_parameter.cache.max_val_age == 1
    assert local_parameter._get_count == 0
    assert local_parameter.cache.get() == value
    assert local_parameter._get_count == 0
    # now fake the time stamp so get should be triggered
    set_time = start - timedelta(seconds=10)
    local_parameter.cache._update_with(
        value=value, raw_value=value, timestamp=set_time)
    # now that ts < max_val_age calling get_latest should update the time
    assert local_parameter.cache.timestamp == set_time
    assert local_parameter.cache.get() == value
    assert local_parameter._get_count == 1
    assert local_parameter.cache.timestamp >= start


def test_no_get_max_val_age():
    """
    Test that cache.get on a parameter with max_val_age set and
    no get cmd raises correctly.
    """
    value = 1
    with pytest.raises(SyntaxError):
        _ = Parameter('test_param', set_cmd=None,
                      get_cmd=False,
                      max_val_age=1, initial_value=value)


def test_no_get_max_val_age_runtime_error(get_if_invalid):
    """
    _BaseParameter does not have a check on creation time that
    no get_cmd is mixed with max_val_age since get_cmd could be added
    in a subclass. Here we create a subclass that does not add a get
    command and also does not implement the check for max_val_age
    """
    value = 1

    class LocalParameter(_BaseParameter):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.set_raw = lambda x: x
            self.set = self._wrap_set(self.set_raw)

    local_parameter = LocalParameter('test_param',
                                     None,
                                     max_val_age=1)
    start = datetime.now()
    set_time = start - timedelta(seconds=10)
    local_parameter.cache._update_with(
        value=value, raw_value=value, timestamp=set_time)

    if get_if_invalid is True:
        with pytest.raises(RuntimeError, match="max_val_age` is not supported"):
            local_parameter.cache.get(get_if_invalid=get_if_invalid)
    elif get_if_invalid == NOT_PASSED:
        with pytest.raises(RuntimeError, match="max_val_age` is not supported"):
            local_parameter.cache.get()
    else:
        assert local_parameter.cache.get(get_if_invalid=get_if_invalid) == 1


def test_no_get_timestamp_none_runtime_error(get_if_invalid):
    """
    Test that a parameter that has never been
    set, cannot be get and does not support
    getting raises a RuntimeError.
    """
    local_parameter = Parameter('test_param', get_cmd=False)

    if get_if_invalid is True:
        with pytest.raises(RuntimeError, match="Value of parameter test_param"):
            local_parameter.cache.get(get_if_invalid=get_if_invalid)
    elif get_if_invalid == NOT_PASSED:
        with pytest.raises(RuntimeError, match="Value of parameter test_param"):
            local_parameter.cache.get()
    else:
        assert local_parameter.cache.get(get_if_invalid=get_if_invalid) is None


def test_latest_dictionary_gets_updated_upon_set_of_memory_parameter():
    p = Parameter('p', set_cmd=None, get_cmd=None)
    assert p.cache._value is None
    assert p.cache._raw_value is None
    assert p.cache.timestamp is None

    p(42)

    assert p.cache._value == 42
    assert p.cache._raw_value == 42
    assert p.cache.timestamp is not None


_P = Parameter


@pytest.mark.parametrize(
    argnames=('p', 'value', 'raw_value'),
    argvalues=(
        (_P('p', set_cmd=None, get_cmd=None), 4, 4),
        (_P('p', set_cmd=False, get_cmd=None), 14, 14),
        (_P('p', set_cmd=None, get_cmd=False), 14, 14),
        (_P('p', set_cmd=None, get_cmd=None, vals=vals.OnOff()), 'on', 'on'),
        (_P('p', set_cmd=None, get_cmd=None, val_mapping={'screw': 1}),
         'screw', 1),
        (_P('p', set_cmd=None, get_cmd=None, set_parser=str, get_parser=int),
         14, '14'),
        (_P('p', set_cmd=None, get_cmd=None, step=7), 14, 14),
        (_P('p', set_cmd=None, get_cmd=None, offset=3), 14, 17),
        (_P('p', set_cmd=None, get_cmd=None, scale=2), 14, 28),
        (_P('p', set_cmd=None, get_cmd=None, offset=-3, scale=2), 14, 25),
    ),
    ids=(
        'with_nothing_extra',
        'without_set_cmd',
        'without_get_cmd',
        'with_on_off_validator',
        'with_val_mapping',
        'with_set_and_parsers',
        'with_step',
        'with_offset',
        'with_scale',
        'with_scale_and_offset',
    )
)
def test_set_latest_works_for_plain_memory_parameter(p, value, raw_value):
    # Set latest value of the parameter
    p.cache.set(value)

    # Assert the latest value and raw_value
    assert p.get_latest() == value
    assert p.raw_value == raw_value

    # Assert latest value and raw_value via private attributes for strictness
    assert p.cache._value == value
    assert p.cache._raw_value == raw_value

    # Now let's get the value of the parameter to ensure that the value that
    # is set above gets picked up from the `_latest` dictionary (due to
    # `get_cmd=None`)

    if not p.gettable:
        assert not hasattr(p, 'get')
        assert p.gettable is False
        return  # finish the test here for non-gettable parameters

    gotten_value = p.get()

    assert gotten_value == value

    # Assert the latest value and raw_value
    assert p.get_latest() == value
    assert p.raw_value == raw_value

    # Assert latest value and raw_value via private attributes for strictness
    assert p.cache._value == value
    assert p.cache._raw_value == raw_value


def test_get_from_cache_marked_invalid():
    param = BetterGettableParam(name="param")
    param.get()
    assert param._get_count == 1

    param.cache.get(get_if_invalid=False)
    assert param._get_count == 1

    param.cache.invalidate()

    param.cache.get(get_if_invalid=True)
    assert param._get_count == 2

    param.cache.invalidate()

    param.cache.get(get_if_invalid=False)
    assert param._get_count == 2

    param._gettable = False

    with pytest.raises(RuntimeError, match="Cannot return cache of a parameter"
                                           " that does not have a get command"
                                           " and has an invalid cache"):
        param.cache.get(get_if_invalid=True)
    assert param._get_count == 2



"""
This file holds the QCoDeS driver for the Galil DMC-41x3 motor controllers.

Colloquially known as the "stepper motors".
"""
from typing import Any, Dict, List, Optional, Tuple

import numpy as np

from qcodes.instrument.base import Instrument
from qcodes.instrument.channel import InstrumentChannel
from qcodes.utils.validators import Enum, Ints, Multiples

try:
    import gclib
except ImportError as e:
    raise ImportError(
        "Cannot find gclib library. Download gclib installer from "
        "https://www.galil.com/sw/pub/all/rn/gclib.html and install Galil "
        "motion controller software for your OS. Afterwards go "
        "to https://www.galil.com/sw/pub/all/doc/gclib/html/python.html and "
        "follow instruction to be able to import gclib package in your "
        "environment."
    ) from e


class GalilMotionController(Instrument):
    """
    Base class for Galil Motion Controller drivers
    """

    def __init__(self, name: str, address: str, **kwargs: Any) -> None:
        """
        Initializes and opens the connection to the Galil Motion Controller

        Args:
            name: name for the instrument
            address: address of the controller
        """
        super().__init__(name=name, **kwargs)
        self.g = gclib.py()
        self._address = address
        self.open()
        self.connect_message()

    def open(self) -> None:
        """
        Open connection to Galil motion controller. This method assumes that
        the initial mapping of Galil motion controller's hardware's
        to an IP address is done using GDK and the IP address is burned in.
        This applies that Motion controller no more requests for an IP address
        and a connection to the Motion controller can be done by the IP
        address burned in.
        """
        self.g.GOpen(self._address + " --direct -s ALL")

    def get_idn(self) -> Dict[str, Optional[str]]:
        """
        Get Galil motion controller hardware information
        """
        data = self.g.GInfo().split(" ")
        idparts: List[Optional[str]] = [
            "Galil Motion Control, Inc.",
            data[1],
            data[4],
            data[3][:-1],
        ]

        return dict(zip(("vendor", "model", "serial", "firmware"), idparts))

    def write_raw(self, cmd: str) -> None:
        """
        Write for Galil motion controller
        """
        self.g.GCommand(cmd + "\r")

    def ask_raw(self, cmd: str) -> str:
        """
        Asks/Reads data from Galil motion controller
        """
        return self.g.GCommand(cmd + "\r")

    def timeout(self, val: int) -> None:
        """
        Sets timeout for the instrument

        Args:
            val: time in milliseconds. -1 disables the timeout
        """
        if val < -1:
            raise RuntimeError("Timeout value cannot be set less than -1")

        self.g.GTimeout(val)

    def close(self) -> None:
        """
        Close connection to the instrument
        """
        self.g.GClose()

    def motion_complete(self, axes: str) -> None:
        """
        Wait for motion to complete for given axes, e.g. ``"A"``
        for axis A, ``"ABC"`` for axes A, B, and C
        """
        self.g.GMotionComplete(axes)


class VectorMode(InstrumentChannel):
    """
    Class to control motors in vector mode
    """

    def __init__(self, parent: "DMC4133Controller", name: str, **kwargs: Any) -> None:
        """
        Initializes the vector mode submodule for the controller

        Args:
            parent: an instance of DMC4133Controller
            name: name of the vector mode plane
        """
        super().__init__(parent, name, **kwargs)
        self._plane = name
        self._vector_position_validator = Ints(
            min_value=-2147483648, max_value=2147483647
        )

        self.add_parameter(
            "coordinate_system",
            get_cmd="CA ?",
            get_parser=self._parse_coordinate_system_active,
            set_cmd="CA {}",
            vals=Enum("S", "T"),
            docstring="activates coordinate system for the motion. Two "
            " coordinate systems are possible with values "
            "'S' and 'T'. All vector mode commands will apply to "
            "the active coordinate system.",
        )

        self.add_parameter(
            "vector_acceleration",
            get_cmd="VA ?",
            get_parser=lambda s: int(float(s)),
            set_cmd="VA {}",
            vals=Multiples(min_value=1024, max_value=1073740800, divisor=1024),
            unit="counts/sec2",
            docstring="sets and gets the defined vector's acceleration",
        )

        self.add_parameter(
            "vector_deceleration",
            get_cmd="VD ?",
            get_parser=lambda s: int(float(s)),
            set_cmd="VD {}",
            vals=Multiples(min_value=1024, max_value=1073740800, divisor=1024),
            unit="counts/sec2",
            docstring="sets and gets the defined vector's deceleration",
        )

        self.add_parameter(
            "vector_speed",
            get_cmd="VS ?",
            get_parser=lambda s: int(float(s)),
            set_cmd="VS {}",
            vals=Multiples(min_value=2, max_value=15000000, divisor=2),
            unit="counts/sec",
            docstring="sets and gets defined vector's speed",
        )

    @staticmethod
    def _parse_coordinate_system_active(val: str) -> str:
        """
        parses the current active coordinate system
        """
        if int(float(val)):
            return "T"
        else:
            return "S"

    def activate(self) -> None:
        """
        activate plane of motion
        """
        self.write(f"VM {self._plane}")

    def vector_position(self, first_coord: int, second_coord: int) -> None:
        """
        sets the final vector position for the motion considering current
        position as the origin
        """
        self._vector_position_validator.validate(first_coord)
        self._vector_position_validator.validate(second_coord)

        self.write(f"VP {first_coord},{second_coord}")

    def vector_seq_end(self) -> None:
        """
        indicates to the controller that the end of the vector is coming up.
        is required to exit the vector mode gracefully
        """
        self.write("VE")

    def begin_seq(self) -> None:
        """
        begins motion of the motor
        """
        self.write("BG S")

    def after_seq_motion(self) -> None:
        """
        wait till motion ends
        """
        self.write("AM S")

    def clear_sequence(self, coord_sys: str) -> None:
        """
        clears vectors specified in the given coordinate system
        """
        if coord_sys not in ["S", "T"]:
            raise RuntimeError(
                f"possible coordinate systems are 'S' or 'T'. "
                f"you provided the following value instead: "
                f" '{coord_sys}'"
            )

        self.write(f"CS {coord_sys}")


class Motor(InstrumentChannel):
    """
    Class to control a single motor (independent of possible other motors)
    """

    def __init__(self, parent: "DMC4133Controller", name: str, **kwargs: Any) -> None:
        """
        Initializes individual motor submodules

        Args:
            parent: an instance of DMC4133Controller
            name: name of the motor to be controlled
        """
        super().__init__(parent, name, **kwargs)
        self._axis = name

        self.add_parameter(
            "relative_position",
            unit="quadrature counts",
            get_cmd=f"MG _PR{self._axis}",
            get_parser=lambda s: int(float(s)),
            set_cmd=self._set_relative_position,
            vals=Ints(-2147483648, 2147483647),
            docstring="sets relative position for the motor's move",
        )

        self.add_parameter(
            "speed",
            unit="counts/sec",
            get_cmd=f"MG _SP{self._axis}",
            get_parser=lambda s: int(float(s)),
            set_cmd=self._set_speed,
            vals=Multiples(min_value=0, max_value=3000000, divisor=2),
            docstring="speed for motor's motion",
        )

        self.add_parameter(
            "acceleration",
            unit="counts/sec2",
            get_cmd=f"MG _AC{self._axis}",
            get_parser=lambda s: int(float(s)),
            set_cmd=self._set_acceleration,
            vals=Multiples(min_value=1024, max_value=1073740800, divisor=1024),
            docstring="acceleration for motor's motion",
        )

        self.add_parameter(
            "deceleration",
            unit="counts/sec2",
            get_cmd=f"MG _DC{self._axis}",
            get_parser=lambda s: int(float(s)),
            set_cmd=self._set_deceleration,
            vals=Multiples(min_value=1024, max_value=1073740800, divisor=1024),
            docstring="deceleration for motor's motion",
        )

        self.add_parameter(
            "off_when_error_occurs",
            get_cmd=self._get_off_when_error_occurs,
            set_cmd=self._set_off_when_error_occurs,
            val_mapping={
                "disable": 0,
                "enable for position, amp error or abort": 1,
                "enable for hw limit switch": 2,
                "enable for all": 3,
            },
            docstring="enables or disables the motor to "
            "automatically turn off when error occurs",
        )

        self.add_parameter(
            "reverse_sw_limit",
            unit="quadrature counts",
            get_cmd=f"MG _BL{self._axis}",
            get_parser=lambda s: int(float(s)),
            set_cmd=self._set_reverse_sw_limit,
            vals=Ints(-2147483648, 2147483647),
            docstring="can be used to set software reverse limit for the motor."
            " motor motion will stop beyond this limit automatically."
            " default value is -2147483648. this value effectively "
            "disables the reverse limit.",
        )

        self.add_parameter(
            "forward_sw_limit",
            unit="quadrature counts",
            get_cmd=f"MG _FL{self._axis}",
            get_parser=lambda s: int(float(s)),
            set_cmd=self._set_forward_sw_limit,
            vals=Ints(-2147483648, 2147483647),
            docstring="can be used to set software forward limit for the motor."
            " motor motion will stop beyond this limit automatically."
            " default value is 2147483647. this value effectively "
            "disables the forward limit.",
        )

    def _set_reverse_sw_limit(self, val: int) -> None:
        """
        Sets reverse software limit
        """
        self.write(f"BL{self._axis}={val}")

    def _set_forward_sw_limit(self, val: int) -> None:
        """
        Sets forward software limit
        """
        self.write(f"FL{self._axis}={val}")

    def _get_off_when_error_occurs(self) -> int:
        """
        Gets the status if motor is automatically set to turn off when error
        occurs.
        """
        val = self.ask(f"MG _OE{self._axis}")

        return int(val[0])

    def _set_off_when_error_occurs(self, val: int) -> None:
        """
        sets the motor to turn off automatically when the error occurs
        """
        self.write(f"OE{self._axis}={val}")

    def _set_deceleration(self, val: str) -> None:
        """
        set deceleration for the motor's motion
        """
        self.write(f"DC{self._axis}={val}")

    def _set_acceleration(self, val: str) -> None:
        """
        set acceleration for the motor's motion
        """
        self.write(f"AC{self._axis}={val}")

    def _set_speed(self, val: str) -> None:
        """
        sets speed for motor's motion
        """
        self.write(f"SP{self._axis}={val}")

    def _set_relative_position(self, val: str) -> None:
        """
        sets relative position
        """
        self.write(f"PR{self._axis}={val}")

    def off(self) -> None:
        """
        turns motor off
        """
        self.write(f"MO {self._axis}")

    def on_off_status(self) -> str:
        """
        tells motor on off status
        """
        val = self.ask(f"MG _MO{self._axis}")
        if val[0] == "1":
            return "off"
        else:
            return "on"

    def servo_here(self) -> None:
        """
        servo at the motor
        """
        self.write(f"SH {self._axis}")

    def begin(self) -> None:
        """
        begins motion of the motor (returns immediately)
        """
        self.write(f"BG {self._axis}")

    def is_in_motion(self) -> int:
        """
        checks if the motor is in motion or not. return 1, if motor is in
        motion otherwise 0
        """
        return int(float(self.ask(f"MG _BG{self._axis}")))

    def wait_till_motor_motion_complete(self) -> None:
        """
        wait for motion on the motor to complete
        """
        try:
            while self.is_in_motion():
                pass
        except KeyboardInterrupt:
            self.root_instrument.abort()
            self.off()

    def error_magnitude(self) -> float:
        """
        gives the magnitude of error, in quadrature counts, for this motor
        a count is directly proportional to the micro-stepping
        resolution of the stepper drive.
        """
        return float(self.ask(f"QS{self._axis}=?"))


class DMC4133Controller(GalilMotionController):
    """
    Driver for Galil DMC-4133 Controller
    """

    def __init__(self, name: str, address: str, **kwargs: Any) -> None:
        """
        Initializes the DMC4133Controller class

        Args:
            name: name for the instance
            address: address of the controller burned in
        """
        super().__init__(name=name, address=address, **kwargs)

        self.add_parameter(
            "position_format_decimals",
            get_cmd=None,
            set_cmd="PF 10.{}",
            vals=Ints(0, 4),
            docstring="sets number of decimals in the format " "of the position",
        )

        self.add_parameter(
            "absolute_position",
            get_cmd=self._get_absolute_position,
            set_cmd=None,
            unit="quadrature counts",
            docstring="gets absolute position of the motors " "from the set origin",
        )

        self.add_parameter(
            "wait",
            get_cmd=None,
            set_cmd="WT {}",
            unit="ms",
            vals=Multiples(min_value=2, max_value=2147483646, divisor=2),
            docstring="controller will wait for the amount of "
            "time specified before executing the next "
            "command",
        )

        self._set_default_update_time()
        self.add_submodule("motor_a", Motor(self, "A"))
        self.add_submodule("motor_b", Motor(self, "B"))
        self.add_submodule("motor_c", Motor(self, "C"))
        self.add_submodule("plane_ab", VectorMode(self, "AB"))
        self.add_submodule("plane_bc", VectorMode(self, "BC"))
        self.add_submodule("plane_ac", VectorMode(self, "AC"))

    def _set_default_update_time(self) -> None:
        """
        sets sampling period to default value of 1000. sampling period affects
        the AC, AS, AT, DC, FA, FV, HV, JG, KP, NB, NF, NZ, PL, SD, SP, VA,
        VD, VS, WT commands.
        """
        self.write("TM 1000")

    def _get_absolute_position(self) -> Dict[str, int]:
        """
        gets absolution position of the motors from the defined origin
        """
        result = dict()
        data = self.ask("PA ?,?,?").split(" ")
        result["A"] = int(data[0][:-1])
        result["B"] = int(data[1][:-1])
        result["C"] = int(data[2])

        return result

    def end_program(self) -> None:
        """
        ends the program
        """
        self.write("EN")

    def define_position_as_origin(self) -> None:
        """
        defines current motors position as origin
        """
        self.write("DP 0,0,0")

    def tell_error(self) -> str:
        """
        reads error
        """
        return self.ask("TC1")

    def stop(self) -> None:
        """
        stop the motion of all motors
        """
        self.write("ST")

    def abort(self) -> None:
        """
        aborts motion and the program operation
        """
        self.write("AB")

    def motors_off(self) -> None:
        """
        turn all motors off
        """
        self.write("MO")

    def begin_motors(self) -> None:
        """
        begin motion of all motors simultaneously
        """
        self.write("BG")

    def wait_till_motion_complete(self) -> None:
        """
        this method waits for the motion on all motors to complete
        """
        try:
            while (
                int(float(self.ask("MG _BGA")))
                or int(float(self.ask("MG _BGB")))
                or int(float(self.ask("MG _BGC")))
            ):
                pass
        except KeyboardInterrupt:
            self.abort()
            self.motors_off()


class Arm:
    """
    Module to control probe arm. It is assumed that the chip to be probed has
    one or more rows with each row having one or more pads. Design of the
    chip to be probed should resemble the following::

        |--------------------------|
        |--------------------------|
        |--------------------------|
        |--------------------------|
        |--------------------------|

    Needle head for the arm in assumed to have rows of needles which is a
    divisor of the number of rows in the chip and number of needles in each
    row of the needle head is a divisor for the number of pads in each row
    of the chip.
    """

    def __init__(self, controller: DMC4133Controller) -> None:
        """
        Initializes the arm class

        Args:
            controller: an instance of DMC4133Controller
        """
        self.controller = controller

        # initialization (all these points will have values in quadrature
        # counts)
        self._left_bottom_position: Optional[Tuple[int, int, int]] = None
        self._left_top_position: Optional[Tuple[int, int, int]] = None
        self._right_top_position: Optional[Tuple[int, int, int]] = None

        # motion directions (all these values are in quadrature counts)
        self._a: np.ndarray  # right_top - left_bottom
        self._b: np.ndarray  # left_top - left_bottom
        self._c: np.ndarray  # right_top - left_top
        self._n: np.ndarray
        self.norm_a: float
        self.norm_b: float
        self.norm_c: float

        # eqn of the chip plane (in quadrature counts)
        self._plane_eqn: np.ndarray

        # current vars
        self._current_row: Optional[int] = None
        self._current_pad: Optional[int] = None

        # chip details
        self.rows: int
        self.pads: int
        self.inter_row_distance: float  # in quadrature counts
        self.inter_pad_distance: float  # in quadrature counts

        # arm kinematics (in the units of quadrature counts)
        self._speed: int
        self._acceleration: int
        self._deceleration: int

        self._arm_pick_up_distance: int

        self._target: np.ndarray

    @property
    def current_row(self) -> Optional[int]:
        return self._current_row

    @property
    def current_pad(self) -> Optional[int]:
        return self._current_pad

    @property
    def left_bottom_position(self) -> Optional[Tuple[int, int, int]]:
        return self._left_bottom_position

    @property
    def left_top_position(self) -> Optional[Tuple[int, int, int]]:
        return self._left_top_position

    @property
    def right_top_position(self) -> Optional[Tuple[int, int, int]]:
        return self._right_top_position

    @property
    def arm_pick_up_distance(self) -> int:
        return self._arm_pick_up_distance

    @property
    def speed(self) -> int:
        return self._speed

    @property
    def acceleration(self) -> int:
        return self._acceleration

    @property
    def deceleration(self) -> int:
        return self._deceleration

    def set_arm_kinematics(
        self, speed: int = 100, acceleration: int = 2048, deceleration: int = 2048
    ) -> None:
        """
        sets the arm kinematics values for speed, acceleration and
        deceleration in micro meters/sec and micro meters/sec^2 respectively
        """

        if acceleration % 256 != 0 or deceleration % 256 != 0:
            raise RuntimeError(
                "Acceleration and deceleration must be a multiple of 256. "
                "Units are micro meters/sec^2"
            )

        self._speed = _convert_micro_meter_to_quadrature_counts(speed)
        self._acceleration = _convert_micro_meter_to_quadrature_counts(acceleration)
        self._deceleration = _convert_micro_meter_to_quadrature_counts(deceleration)

    def set_pick_up_distance(self, distance: float = 3000) -> None:
        """
        sets pick up distance in micrometers for the arm
        """

        self._arm_pick_up_distance = _convert_micro_meter_to_quadrature_counts(distance)

    def set_left_bottom_position(self) -> None:

        pos = self.controller.absolute_position()
        self._left_bottom_position = (pos["A"], pos["B"], pos["C"])

        self._calculate_ortho_vector()

    def set_left_top_position(self) -> None:

        pos = self.controller.absolute_position()
        self._left_top_position = (pos["A"], pos["B"], pos["C"])

        self._calculate_ortho_vector()

    def set_right_top_position(self) -> None:

        pos = self.controller.absolute_position()
        self._right_top_position = (pos["A"], pos["B"], pos["C"])

        self._calculate_ortho_vector()

    def _calculate_ortho_vector(self) -> None:

        if (
            self._left_bottom_position is None
            or self._left_top_position is None
            or self._right_top_position is None
        ):
            return

        a = np.asarray(self._right_top_position) - np.asarray(
            self._left_bottom_position
        )
        self.norm_a = np.linalg.norm(a)
        self._a = a / self.norm_a

        b = np.asarray(self._left_top_position) - np.asarray(self._left_bottom_position)
        self.norm_b = np.linalg.norm(b)
        self._b = b / self.norm_b

        c = np.asarray(self._right_top_position) - np.asarray(self._left_top_position)
        self.norm_c = np.linalg.norm(c)
        self._c = c / self.norm_c

        n = np.cross(self._a, self._b)
        norm_n = np.linalg.norm(n)
        self._n = n / norm_n

        intercept = np.array(sum(-1 * self._n * self._left_bottom_position))
        self._plane_eqn = np.append(self._n, intercept)

    def move_motor_a_by(self, distance: float) -> None:
        """
        Moves motor A by distance given in micro meters
        """
        a = self.controller.motor_a

        d = _convert_micro_meter_to_quadrature_counts(distance)

        a.relative_position(d)
        a.speed(self._speed)
        a.acceleration(self._acceleration)
        a.deceleration(self._deceleration)
        a.servo_here()
        a.begin()
        a.wait_till_motor_motion_complete()

    def move_motor_b_by(self, distance: float) -> None:
        """
        Moves motor B by distance given in micro meters
        """
        b = self.controller.motor_b

        d = _convert_micro_meter_to_quadrature_counts(distance)

        b.relative_position(d)
        b.speed(self._speed)
        b.acceleration(self._acceleration)
        b.deceleration(self._deceleration)
        b.servo_here()
        b.begin()
        b.wait_till_motor_motion_complete()

    def move_motor_c_by(self, distance: float) -> None:
        """
        Moves motor B by distance given in micro meters
        """
        c = self.controller.motor_c

        d = _convert_micro_meter_to_quadrature_counts(distance)

        c.relative_position(d)
        c.speed(self._speed)
        c.acceleration(self._acceleration)
        c.deceleration(self._deceleration)
        c.servo_here()
        c.begin()
        c.wait_till_motor_motion_complete()

    def _setup_motion(self, rel_vec: np.ndarray, d: float, speed: float) -> None:
        """
        sets up motion parameters. all arguments have units in quadrature counts
        """

        pos = self.controller.absolute_position()

        a = int(np.floor(rel_vec[0] * d))
        b = int(np.floor(rel_vec[1] * d))
        c = int(np.floor(rel_vec[2] * d))

        self._target = np.array([pos["A"] + a, pos["B"] + b, pos["C"] + c, 1])

        if np.dot(self._plane_eqn, self._target) < 0:
            flag = 1
            for idx, coord in enumerate([a, b, c]):
                temp1 = coord - 1
                temp2 = coord + 1
                if idx == 0:
                    target1 = np.array(
                        [pos["A"] + temp1, pos["B"] + b, pos["C"] + c, 1]
                    )
                    target2 = np.array(
                        [pos["A"] + temp2, pos["B"] + b, pos["C"] + c, 1]
                    )
                elif idx == 1:
                    target1 = np.array(
                        [pos["A"] + a, pos["B"] + temp1, pos["C"] + c, 1]
                    )
                    target2 = np.array(
                        [pos["A"] + a, pos["B"] + temp2, pos["C"] + c, 1]
                    )
                else:
                    target1 = np.array(
                        [pos["A"] + a, pos["B"] + b, pos["C"] + temp1, 1]
                    )
                    target2 = np.array(
                        [pos["A"] + a, pos["B"] + b, pos["C"] + temp2, 1]
                    )

                if np.dot(self._plane_eqn, target1) >= 0:
                    flag = 0
                    if idx == 0:
                        a = temp1
                    elif idx == 1:
                        b = temp1
                    else:
                        c = temp1
                    break

                if np.dot(self._plane_eqn, target2) >= 0:
                    flag = 0
                    if idx == 0:
                        a = temp2
                    elif idx == 1:
                        b = temp2
                    else:
                        c = temp2
                    break

            if flag:
                raise RuntimeError(
                    f"Cannot move to {self._target[:3]}. Target location is "
                    f"below chip plane."
                )

        sp_a = int(np.floor(abs(rel_vec[0]) * speed))
        if sp_a % 2 != 0:
            sp_a += 1

        sp_b = int(np.floor(abs(rel_vec[1]) * speed))
        if sp_b % 2 != 0:
            sp_b += 1

        sp_c = int(np.floor(abs(rel_vec[2]) * speed))
        if sp_c % 2 != 0:
            sp_c += 1

        acc_a = _calculate_vector_component(vec=rel_vec[0], val=self._acceleration)
        acc_b = _calculate_vector_component(vec=rel_vec[1], val=self._acceleration)
        acc_c = _calculate_vector_component(vec=rel_vec[2], val=self._acceleration)

        dec_a = _calculate_vector_component(vec=rel_vec[0], val=self._deceleration)
        dec_b = _calculate_vector_component(vec=rel_vec[1], val=self._deceleration)
        dec_c = _calculate_vector_component(vec=rel_vec[2], val=self._deceleration)

        motor_a = self.controller.motor_a
        motor_b = self.controller.motor_b
        motor_c = self.controller.motor_c

        motor_a.relative_position(a)
        motor_a.speed(sp_a)
        motor_a.acceleration(acc_a)
        motor_a.deceleration(dec_a)
        motor_a.servo_here()

        motor_b.relative_position(b)
        motor_b.speed(sp_b)
        motor_b.acceleration(acc_b)
        motor_b.deceleration(dec_b)
        motor_b.servo_here()

        motor_c.relative_position(c)
        motor_c.speed(sp_c)
        motor_c.acceleration(acc_c)
        motor_c.deceleration(dec_c)
        motor_c.servo_here()

    def _move(self) -> None:
        self.controller.begin_motors()
        self.controller.wait_till_motion_complete()

    def _pick_up(self) -> None:

        self.move_motor_c_by(distance=-20)
        self._setup_motion(
            rel_vec=self._n, d=self._arm_pick_up_distance, speed=self._speed
        )
        self._move()

    def _put_down(self) -> None:

        motion_vec = -1 * self._n

        pos = self.controller.absolute_position()
        x1 = pos["A"]
        y1 = pos["B"]
        z1 = pos["C"]
        current = np.array([x1, y1, z1, 1])

        denominator = np.sqrt(
            self._plane_eqn[0] ** 2 + self._plane_eqn[1] ** 2 + self._plane_eqn[2] ** 2
        )
        d = abs(sum(self._plane_eqn * current)) / denominator

        self._setup_motion(rel_vec=motion_vec, d=d, speed=self._speed)
        self._move()

    def move_towards_left_bottom_position(self) -> None:

        self._pick_up()

        motion_vec = -1 * self._a
        self._setup_motion(rel_vec=motion_vec, d=self.norm_a, speed=self._speed)
        self._move()
        self._put_down()

        self._current_row = 1
        self._current_pad = 1

    def move_to_next_row(self) -> None:

        if self._current_row is None or self._current_pad is None:
            raise RuntimeError("Current position unknown.")

        if self._current_row == self.rows:
            raise RuntimeError("Cannot move further")

        self._pick_up()

        self._setup_motion(
            rel_vec=self._b, d=self.inter_row_distance, speed=self._speed
        )
        self._move()

        self._put_down()

        self._current_row = self._current_row + 1

    def move_to_begin_row_pad_from_end_row_last_pad(self) -> None:

        if self._current_row is None or self._current_pad is None:
            raise RuntimeError("Current position unknown.")

        if self._current_pad == self.pads:
            raise RuntimeError("Cannot move further")

        motion_vec = -1 * self._b * self.norm_b + self._c * self.inter_pad_distance
        norm = np.linalg.norm(motion_vec)
        motion_vec_cap = motion_vec / norm

        self._pick_up()

        self._setup_motion(rel_vec=motion_vec_cap, d=norm, speed=self._speed)
        self._move()

        self._put_down()

        self._current_row = 1
        self._current_pad = self._current_pad + 1

    def move_to_row(self, num: int) -> None:

        if num < 1 or num > self.rows:
            raise RuntimeError(
                f"Row num: {num} is out of range. Row numbers start from 1 "
                f"and max number of rows is {self.rows}."
            )

        sign = 1
        assert self._current_row is not None
        d = (num - self._current_row) * self.inter_row_distance

        if d == 0:
            raise RuntimeError(
                f"You are at the row where you want to be. Current row number "
                f"is {self._current_row}."
            )
        if d < 0:
            sign = -1
            d = abs(d)

        self._pick_up()

        self._setup_motion(rel_vec=sign * self._b, d=d, speed=self._speed)
        self._move()

        self._put_down()

        self._current_row = num

    def move_to_pad(self, num: int) -> None:

        if num < 1 or num > self.pads:
            raise RuntimeError(
                f"Pad num: {num} is out of range. Pad number start from 1 "
                f"and max number is {self.pads}."
            )

        sign = 1
        assert self._current_pad is not None
        d = (num - self._current_pad) * self.inter_pad_distance

        if d == 0:
            raise RuntimeError(
                f"You are at the pad where you want to be. Current pad "
                f"number is {self._current_pad}."
            )
        if d < 0:
            sign = -1
            d = abs(d)

        self._pick_up()

        self._setup_motion(rel_vec=sign * self._c, d=d, speed=self._speed)
        self._move()

        self._put_down()

        self._current_pad = num

    def set_motor_a_forward_limit(self) -> None:

        pos = self.controller.absolute_position()
        self.controller.motor_a.forward_sw_limit(pos["A"])

    def set_motor_a_reverse_limit(self) -> None:

        pos = self.controller.absolute_position()
        self.controller.motor_a.reverse_sw_limit(pos["A"])

    def set_motor_b_forward_limit(self) -> None:

        pos = self.controller.absolute_position()
        self.controller.motor_b.forward_sw_limit(pos["B"])

    def set_motor_b_reverse_limit(self) -> None:

        pos = self.controller.absolute_position()
        self.controller.motor_b.reverse_sw_limit(pos["B"])

    def set_motor_c_forward_limit(self) -> None:

        pos = self.controller.absolute_position()
        self.controller.motor_c.forward_sw_limit(pos["C"])

    def set_motor_c_reverse_limit(self) -> None:

        pos = self.controller.absolute_position()
        self.controller.motor_c.reverse_sw_limit(pos["C"])


def _convert_micro_meter_to_quadrature_counts(val: float) -> int:

    return int(20 * val)


def _calculate_vector_component(vec: float, val: int) -> int:
    return_val = int(np.floor(abs(vec) * val))
    return_val = return_val + 1024
    rem = return_val % 1024
    return_val = return_val - rem

    assert return_val % 1024 == 0

    return return_val



import pytest

from qcodes.instrument.parameter import ArrayParameter
from .conftest import blank_instruments, named_instrument


class SimpleArrayParam(ArrayParameter):
    def __init__(self, return_val, *args, **kwargs):
        self._return_val = return_val
        self._get_count = 0
        super().__init__(*args, **kwargs)

    def get_raw(self):
        self._get_count += 1
        return self._return_val


class SettableArray(SimpleArrayParam):
    # this is not allowed - just created to raise an error in the test below
    def set_raw(self, value):
        self.v = value


def test_default_attributes():
    name = 'array_param'
    shape = (2, 3)
    p = SimpleArrayParam([[1, 2, 3], [4, 5, 6]], name, shape)

    assert p.name == name
    assert p.shape == shape

    assert p.label == name
    assert p.unit == ''
    assert p.setpoints is None
    assert p.setpoint_names is None
    assert p.setpoint_labels is None

    assert str(p) == name

    assert p._get_count == 0
    snap = p.snapshot(update=True)
    assert p._get_count == 0
    snap_expected = {
        'name': name,
        'label': name,
        'unit': ''
    }
    for k, v in snap_expected.items():
        assert snap[k] == v
    assert 'value' not in snap
    assert 'raw_value' not in snap
    assert snap['ts'] is None

    assert name in p.__doc__


def test_explicit_attributes():
    name = 'tiny_array'
    shape = (2,)
    label = 'it takes two to tango'
    unit = 'steps'
    setpoints = [(0, 1)]
    setpoint_names = ['sp_index']
    setpoint_labels = ['Setpoint Label']
    docstring = 'Whats up Doc?'
    metadata = {'size': 2}
    p = SimpleArrayParam([6, 7], name, shape, label=label, unit=unit,
                         setpoints=setpoints,
                         setpoint_names=setpoint_names,
                         setpoint_labels=setpoint_labels,
                         docstring=docstring, snapshot_value=True,
                         metadata=metadata)

    assert p.name == name
    assert p.shape == shape
    assert p.label == label
    assert p.unit == unit
    assert p.setpoints == setpoints
    assert p.setpoint_names == setpoint_names
    assert p.setpoint_full_names == setpoint_names
    assert p.setpoint_labels == setpoint_labels

    assert p._get_count == 0
    snap = p.snapshot(update=True)
    assert p._get_count == 1
    snap_expected = {
        'name': name,
        'label': label,
        'unit': unit,
        'setpoint_names': setpoint_names,
        'setpoint_labels': setpoint_labels,
        'metadata': metadata,
        'value': [6, 7],
        'raw_value': [6, 7]
    }
    for k, v in snap_expected.items():
        assert snap[k] == v
    assert snap['ts'] is not None

    assert name in p.__doc__
    assert docstring in p.__doc__


def test_has_set_get():
    name = 'array_param'
    shape = (3,)
    with pytest.raises(AttributeError):
        ArrayParameter(name, shape)

    p = SimpleArrayParam([1, 2, 3], name, shape)

    assert hasattr(p, 'get')
    assert p.gettable
    assert not hasattr(p, 'set')
    assert not p.settable

    # Yet, it's possible to set the cached value
    p.cache.set([6, 7, 8])
    assert p.get_latest() == [6, 7, 8]
    # However, due to the implementation of this ``SimpleArrayParam``
    # test parameter it's ``get`` call will return the originally passed
    # list
    assert p.get() == [1, 2, 3]
    assert p.get_latest() == [1, 2, 3]

    with pytest.raises(AttributeError):
        SettableArray([1, 2, 3], name, shape)


def test_full_name():
    # three cases where only name gets used for full_name
    for instrument in blank_instruments:
        p = SimpleArrayParam([6, 7], 'fred', (2,),
                             setpoint_names=('barney',))
        p._instrument = instrument
        assert str(p) == 'fred'
        assert p.setpoint_full_names == ('barney',)

    # and then an instrument that really has a name
    p = SimpleArrayParam([6, 7], 'wilma', (2,),
                         setpoint_names=('betty',))
    p._instrument = named_instrument
    assert str(p) == 'astro_wilma'
    assert p.setpoint_full_names == ('astro_betty',)

    # and with a 2d parameter to test mixed setpoint_names
    p = SimpleArrayParam([[6, 7, 8], [1, 2, 3]], 'wilma', (3, 2),
                         setpoint_names=('betty', None))
    p._instrument = named_instrument
    assert p.setpoint_full_names == ('astro_betty', None)


@pytest.mark.parametrize("constructor", [
    {'shape': [[3]]},  # not a depth-1 sequence
    {'shape': [3], 'setpoints': [1, 2, 3]},  # should be [[1, 2, 3]]
    {'shape': [3], 'setpoint_names': 'index'},  # should be ['index']
    {'shape': [3], 'setpoint_labels': 'the index'},  # ['the index']
    {'shape': [3], 'setpoint_names': [None, 'index2']}
])
def test_constructor_errors(constructor):
    with pytest.raises(ValueError):
        SimpleArrayParam([1, 2, 3], 'p', **constructor)



"""
This module contains functions which implement conversion between different
(neighbouring) versions of RunDescriber.



"""
from typing import Dict, List

from ..dependencies import InterDependencies_
from ..param_spec import ParamSpec, ParamSpecBase
from .rundescribertypes import (RunDescriberV0Dict, RunDescriberV1Dict,
                                RunDescriberV2Dict, RunDescriberV3Dict)
from .v0 import InterDependencies


def old_to_new(idps: InterDependencies) -> InterDependencies_:
    """
    Create a new InterDependencies_ object (new style) from an existing
    InterDependencies object (old style). Leaves the original object unchanged.
    Incidentally, this function can serve as a validator of the original object
    """
    namedict: Dict[str, ParamSpec] = {ps.name: ps for ps in idps.paramspecs}

    dependencies = {}
    inferences = {}
    standalones_mut = []
    root_paramspecs: List[ParamSpecBase] = []

    for ps in idps.paramspecs:
        deps = tuple(namedict[n].base_version() for n in ps.depends_on_)
        inffs = tuple(namedict[n].base_version() for n in ps.inferred_from_)
        if len(deps) > 0:
            dependencies.update({ps.base_version(): deps})
            root_paramspecs += list(deps)
        if len(inffs) > 0:
            inferences.update({ps.base_version(): inffs})
            root_paramspecs += list(inffs)
        if len(deps) == len(inffs) == 0:
            standalones_mut.append(ps.base_version())

    standalones = tuple(set(standalones_mut).difference(set(root_paramspecs)))

    idps_ = InterDependencies_(dependencies=dependencies,
                               inferences=inferences,
                               standalones=standalones)
    return idps_


def new_to_old(idps: InterDependencies_) -> InterDependencies:
    """
    Create a new InterDependencies object (old style) from an existing
    InterDependencies_ object (new style). Leaves the original object
    unchanged. Only meant to be used for ensuring backwards-compatibility
    until we update sqlite module to forget about ParamSpecs
    """

    paramspecs: Dict[str, ParamSpec] = {}

    # first the independent parameters
    for indeps in idps.dependencies.values():
        for indep in indeps:
            paramspecs.update({indep.name: ParamSpec(name=indep.name,
                                                     paramtype=indep.type,
                                                     label=indep.label,
                                                     unit=indep.unit)})

    for inffs in idps.inferences.values():
        for inff in inffs:
            paramspecs.update({inff.name: ParamSpec(name=inff.name,
                                                    paramtype=inff.type,
                                                    label=inff.label,
                                                    unit=inff.unit)})

    for ps_base in idps._paramspec_to_id.keys():
        paramspecs.update({ps_base.name: ParamSpec(name=ps_base.name,
                                                   paramtype=ps_base.type,
                                                   label=ps_base.label,
                                                   unit=ps_base.unit)})

    for ps, indeps in idps.dependencies.items():
        for indep in indeps:
            paramspecs[ps.name]._depends_on.append(indep.name)
    for ps, inffs in idps.inferences.items():
        for inff in inffs:
            paramspecs[ps.name]._inferred_from.append(inff.name)

    return InterDependencies(*tuple(paramspecs.values()))


def v0_to_v1(old: RunDescriberV0Dict) -> RunDescriberV1Dict:
    """
    Convert a v0 RunDescriber Dict to a v1 RunDescriber Dict
    """
    old_idps = InterDependencies._from_dict(old["interdependencies"])
    new_idps_dict = old_to_new(old_idps)._to_dict()
    return RunDescriberV1Dict(version=1, interdependencies=new_idps_dict)


def v1_to_v2(old: RunDescriberV1Dict) -> RunDescriberV2Dict:
    """
    Convert a v1 RunDescriber Dict to a v2 RunDescriber Dict
    """
    interdeps_dict = old['interdependencies']
    interdeps_ = InterDependencies_._from_dict(interdeps_dict)
    interdepsdict = new_to_old(interdeps_)._to_dict()
    return RunDescriberV2Dict(version=2, interdependencies_=interdeps_dict,
                              interdependencies=interdepsdict)


def v2_to_v3(old: RunDescriberV2Dict) -> RunDescriberV3Dict:
    return RunDescriberV3Dict(version=3,
                              interdependencies=old['interdependencies'],
                              interdependencies_=old['interdependencies_'],
                              shapes=None
                              )


def v0_to_v2(old: RunDescriberV0Dict) -> RunDescriberV2Dict:
    """
    Convert a v0 RunDescriber Dict to a v2 RunDescriber Dict
    """
    return v1_to_v2(v0_to_v1(old))


def v0_to_v3(old: RunDescriberV0Dict) -> RunDescriberV3Dict:
    return v2_to_v3(v0_to_v2(old))


def v1_to_v3(old: RunDescriberV1Dict) -> RunDescriberV3Dict:
    return v2_to_v3(v1_to_v2(old))


def v3_to_v2(new: RunDescriberV3Dict) -> RunDescriberV2Dict:
    return RunDescriberV2Dict(version=2,
                              interdependencies=new['interdependencies'],
                              interdependencies_=new['interdependencies_'],
                              )


def v2_to_v1(new: RunDescriberV2Dict) -> RunDescriberV1Dict:
    """
    Convert a v2 RunDescriber Dict to a v1 RunDescriber Dict
    """
    rundescriberdictv1 = RunDescriberV1Dict(
        version=1,
        interdependencies=new['interdependencies_']
    )
    return rundescriberdictv1


def v1_to_v0(new: RunDescriberV1Dict) -> RunDescriberV0Dict:
    """
    Convert a v1 RunDescriber Dict to a v0 RunDescriber Dict
    """
    interdeps_dict = new['interdependencies']
    interdeps_ = InterDependencies_._from_dict(interdeps_dict)
    interdepsdict = new_to_old(interdeps_)._to_dict()
    rundescriberv0dict = RunDescriberV0Dict(version=0,
                                            interdependencies=interdepsdict)
    return rundescriberv0dict


def v3_to_v1(new: RunDescriberV3Dict) -> RunDescriberV1Dict:
    return v2_to_v1(v3_to_v2(new))


def v2_to_v0(new: RunDescriberV2Dict) -> RunDescriberV0Dict:
    """
    Convert a v2 RunDescriber Dict to a v0 RunDescriber Dict
    """
    return v1_to_v0(v2_to_v1(new))


def v3_to_v0(new: RunDescriberV3Dict) -> RunDescriberV0Dict:
    return v1_to_v0(v3_to_v1(new))
